"""Hyperparameter optimization for model training."""

import json
import math
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

# Build Morgan fingerprints
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit import DataStructs


import numpy as np
import pandas as pd

from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error, r2_score
from scipy.stats import pearsonr, spearmanr
from sklearn.model_selection import KFold, cross_val_score
from sklearn.svm import SVR

from dta_gnn.features import calculate_morgan_fingerprints


def _resolve_stratified_folds(y: np.ndarray, requested_folds: int) -> int:
    """Resolve a safe number of CV folds for StratifiedKFold.

    Stratified K-fold requires at least `n_splits` samples in every class.
    """
    if requested_folds < 2:
        raise ValueError("cv_folds must be >= 2")

    y = np.asarray(y)
    if y.size < 2:
        raise ValueError(
            "Not enough samples for cross-validation. "
            "Build a larger dataset (train+val must have at least 2 rows)."
        )

    unique, counts = np.unique(y, return_counts=True)
    if unique.size < 2:
        raise ValueError(
            "Hyperparameter optimization needs at least two label classes. "
            "Your current run has only one class in train/val."
        )

    min_class = int(counts.min())
    folds = min(int(requested_folds), int(y.size), min_class)
    if folds < 2:
        raise ValueError(
            "Not enough samples per class for cross-validation. "
            f"Minimum class count is {min_class}; need at least 2 per class."
        )
    return folds




def _require_wandb():
    """Ensure Weights & Biases is installed."""
    try:
        import wandb

        return wandb
    except ImportError as e:
        raise ImportError(
            "wandb is not installed. Install with: `pip install 'dta_gnn[wandb]'`"
        ) from e


@dataclass
class HyperoptConfig:
    """Configuration for hyperparameter optimization."""

    model_type: Literal["RandomForest", "SVR", "GNN"]
    n_trials: int = 20
    n_jobs: int = 1
    sampler_seed: int = 42

    # RF parameters to optimize
    rf_optimize_n_estimators: bool = False
    rf_n_estimators_min: int = 50
    rf_n_estimators_max: int = 500

    rf_optimize_max_depth: bool = False
    rf_max_depth_min: int = 5
    rf_max_depth_max: int = 50

    rf_optimize_min_samples_split: bool = False
    rf_min_samples_split_min: int = 2
    rf_min_samples_split_max: int = 20

    # SVR parameters to optimize (regression only)
    svr_optimize_C: bool = False
    svr_C_min: float = 0.1
    svr_C_max: float = 100.0
    svr_C_default: float = 10.0

    svr_optimize_epsilon: bool = False
    svr_epsilon_min: float = 0.01
    svr_epsilon_max: float = 0.2
    svr_epsilon_default: float = 0.1

    svr_optimize_kernel: bool = False
    svr_kernel_choices: list[str] = None  # type: ignore[assignment]
    svr_kernel_default: str = "rbf"

    # GNN parameters to optimize
    architecture: Literal[
        "gin",
        "gcn",
        "gat",
        "sage",
        "pna",
        "transformer",
        "tag",
        "arma",
        "cheb",
        "supergat",
    ] = "gin"

    optimize_epochs: bool = False
    epochs_min: int = 5
    epochs_max: int = 50
    epochs_default: int = 20

    optimize_lr: bool = False
    lr_min: float = 0.00001
    lr_max: float = 0.01

    optimize_batch_size: bool = False
    batch_size_min: int = 16
    batch_size_max: int = 256
    batch_size_default: int = 64

    optimize_embedding_dim: bool = False
    embedding_dim_min: int = 32
    embedding_dim_max: int = 512
    embedding_dim_default: int = 128

    optimize_hidden_dim: bool = False
    hidden_dim_min: int = 32
    hidden_dim_max: int = 512
    hidden_dim_default: int = 128

    optimize_num_layers: bool = False
    num_layers_min: int = 1
    num_layers_max: int = 5
    num_layers_default: int = 3

    # Extra GNN architecture knobs (GIN-only today)
    optimize_dropout: bool = False
    dropout_min: float = 0.0
    dropout_max: float = 0.6
    dropout_default: float = 0.1

    optimize_pooling: bool = False
    pooling_choices: list[str] = None  # type: ignore[assignment]
    pooling_default: str = "add"

    optimize_residual: bool = False
    residual_default: bool = False

    optimize_head_mlp_layers: bool = False
    head_mlp_layers_min: int = 1
    head_mlp_layers_max: int = 4
    head_mlp_layers_default: int = 2

    optimize_gin_conv_mlp_layers: bool = False
    gin_conv_mlp_layers_min: int = 1
    gin_conv_mlp_layers_max: int = 4
    gin_conv_mlp_layers_default: int = 2

    # GIN-specific epsilon parameters
    optimize_gin_train_eps: bool = False
    gin_train_eps_default: bool = False
    optimize_gin_eps: bool = False
    gin_eps_min: float = 0.0
    gin_eps_max: float = 1.0
    gin_eps_default: float = 0.0

    # GAT-specific
    optimize_gat_heads: bool = False
    gat_heads_min: int = 1
    gat_heads_max: int = 8
    gat_heads_default: int = 4

    # GraphSAGE-specific
    optimize_sage_aggr: bool = False
    sage_aggr_choices: list[str] = None  # type: ignore[assignment]
    sage_aggr_default: str = "mean"

    # Transformer-specific
    optimize_transformer_heads: bool = False
    transformer_heads_min: int = 1
    transformer_heads_max: int = 8
    transformer_heads_default: int = 4

    # TAG-specific
    optimize_tag_k: bool = False
    tag_k_min: int = 1
    tag_k_max: int = 5
    tag_k_default: int = 2  # Reduced from 3 for better performance (K=2 is often sufficient)

    # ARMA-specific
    optimize_arma_stacks: bool = False
    arma_num_stacks_min: int = 1
    arma_num_stacks_max: int = 3
    arma_num_stacks_default: int = 1
    optimize_arma_layers: bool = False
    arma_num_layers_min: int = 1
    arma_num_layers_max: int = 3
    arma_num_layers_default: int = 1

    # Cheb-specific
    optimize_cheb_k: bool = False
    cheb_k_min: int = 1
    cheb_k_max: int = 5
    cheb_k_default: int = 2

    # SuperGAT-specific
    optimize_supergat_heads: bool = False
    supergat_heads_min: int = 1
    supergat_heads_max: int = 8
    supergat_heads_default: int = 4
    optimize_supergat_attention_type: bool = False
    supergat_attention_type_choices: list[str] = None  # type: ignore[assignment]
    supergat_attention_type_default: str = "MX"  # "MX" (Mixed) or "SD" (Self-Distillation)

    # Device for GNN training
    device: str | None = None  # "auto", "mps", "cuda", "cpu", or None (auto)


@dataclass
class HyperoptResult:
    """Result from hyperparameter optimization."""

    run_dir: Path
    best_params: dict
    best_value: float
    best_trial_number: int
    n_trials: int
    study_path: str
    best_params_path: str
    strategy: Literal["holdout-val", "cv"]
    cv_folds_used: Optional[int]


def _score_from_gnn_metrics(task_type: str, metrics: dict) -> float:
    """Single scalar score to maximize for regression."""
    val = (metrics or {}).get("splits", {}).get("val", {}) or {}
    # Regression: prefer r2, otherwise maximize -rmse.
    r2 = val.get("r2")
    if r2 is not None:
        try:
            return float(r2)
        except Exception:
            pass
    rmse = val.get("rmse")
    return -float(rmse) if rmse is not None else 0.0


def _infer_task_type_from_metadata_or_labels(
    df: pd.DataFrame,
    meta: dict | None,
) -> Literal["regression"]:
    return "regression"


def _bitstrings_to_numpy(bitstrings: list[str], *, n_bits: int) -> np.ndarray:
    out = np.empty((len(bitstrings), n_bits), dtype=np.uint8)
    for i, s in enumerate(bitstrings):
        b = np.frombuffer(str(s).encode("ascii"), dtype=np.uint8)
        if b.size != n_bits:
            raise ValueError(
                f"Fingerprint length mismatch: expected {n_bits}, got {b.size}."
            )
        out[i, :] = b - 48
    return out


def optimize_random_forest_wandb(
    run_dir: Path,
    *,
    config: HyperoptConfig,
    project: str,
    entity: str | None = None,
    api_key: str | None = None,
    sweep_name: str | None = None,
    radius: int = 2,
    n_bits: int = 2048,
) -> HyperoptResult:
    """Optimize RandomForest hyperparameters using a W&B Bayes sweep.

    Uses:
    - Holdout validation if a val split exists.
    - Otherwise CV (KFold for regression).
    """

    wandb = _require_wandb()

    run_dir = Path(run_dir).resolve()
    dataset_path = run_dir / "dataset.csv"
    compounds_path = run_dir / "compounds.csv"
    metadata_path = run_dir / "metadata.json"

    if not dataset_path.exists() or not compounds_path.exists():
        raise FileNotFoundError(f"Expected {dataset_path} and {compounds_path}")

    df = pd.read_csv(dataset_path)
    compounds = pd.read_csv(compounds_path)

    if "molecule_chembl_id" not in df.columns or "label" not in df.columns:
        raise ValueError("dataset.csv must contain 'molecule_chembl_id' and 'label'.")
    if (
        "molecule_chembl_id" not in compounds.columns
        or "smiles" not in compounds.columns
    ):
        raise ValueError(
            "compounds.csv must contain 'molecule_chembl_id' and 'smiles'."
        )

    meta = None
    try:
        if metadata_path.exists():
            meta = json.loads(metadata_path.read_text())
    except Exception:
        meta = None

    task_type = _infer_task_type_from_metadata_or_labels(df, meta)

    # Compute fingerprints once.


    df_comp = (
        compounds[["molecule_chembl_id", "smiles"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )
    df_feat = calculate_morgan_fingerprints(
        df_comp,
        smiles_col="smiles",
        radius=int(radius),
        n_bits=int(n_bits),
        out_col="morgan_fingerprint",
        drop_failures=True,
    )
    df_feat = (
        df_feat[["molecule_chembl_id", "morgan_fingerprint"]].dropna().drop_duplicates()
    )
    fp_map = dict(
        zip(
            df_feat["molecule_chembl_id"].astype(str),
            df_feat["morgan_fingerprint"].astype(str),
        )
    )

    df2 = df.copy()
    df2["_fp"] = df2["molecule_chembl_id"].astype(str).map(fp_map)
    df2 = df2.dropna(subset=["_fp"]).reset_index(drop=True)
    if df2.empty:
        raise ValueError("No rows left after joining fingerprints to dataset.")

    X = _bitstrings_to_numpy(df2["_fp"].astype(str).tolist(), n_bits=int(n_bits))
    y = df2["label"].to_numpy()

    # Exclude test split for tuning when present.
    if "split" in df2.columns:
        base_mask = df2["split"].astype(str).ne("test").to_numpy()
        has_val = bool((df2.loc[base_mask, "split"].astype(str) == "val").any())
    else:
        base_mask = np.ones(len(df2), dtype=bool)
        has_val = False

    project = (project or "").strip() or "dta_gnn"
    entity = (entity or "").strip() or None
    if api_key and str(api_key).strip():
        wandb.login(key=str(api_key).strip(), relogin=True)

    # Build sweep parameter space only from enabled knobs.
    parameters: dict[str, dict] = {}
    if config.rf_optimize_n_estimators:
        parameters["n_estimators"] = {
            "distribution": "int_uniform",
            "min": int(config.rf_n_estimators_min),
            "max": int(config.rf_n_estimators_max),
        }
    if config.rf_optimize_max_depth:
        parameters["max_depth"] = {
            "distribution": "int_uniform",
            "min": int(config.rf_max_depth_min),
            "max": int(config.rf_max_depth_max),
        }
    if config.rf_optimize_min_samples_split:
        parameters["min_samples_split"] = {
            "distribution": "int_uniform",
            "min": int(config.rf_min_samples_split_min),
            "max": int(config.rf_min_samples_split_max),
        }

    if not parameters:
        raise ValueError(
            "No parameters selected for optimization. "
            "Enable at least one 'Optimize ...' checkbox before running a sweep."
        )

    sweep_config: dict[str, object] = {
        "name": sweep_name or f"dta_gnn_rf_{task_type}",
        "method": "bayes",
        "metric": {"name": "val_score", "goal": "maximize"},
        "parameters": parameters,
    }

    sweep_id = wandb.sweep(sweep=sweep_config, project=project, entity=entity)

    best_score = -math.inf
    best_params: dict[str, object] = {}
    best_trial_number = -1
    trial_counter = {"i": 0}

    def _trial_fn():
        nonlocal best_score, best_params, best_trial_number

        run = wandb.init(project=project, entity=entity, config={})
        trial_idx = int(trial_counter["i"])
        trial_counter["i"] = trial_idx + 1

        sampled = dict(getattr(wandb, "config", {}) or {})

        n_estimators = int(sampled.get("n_estimators", 500))
        max_depth = sampled.get("max_depth", None)
        max_depth = int(max_depth) if max_depth is not None else None
        min_samples_split = int(sampled.get("min_samples_split", 2))

        score: float
        extra_logs: dict[str, object] = {}
           
        model = RandomForestRegressor(
            n_estimators=n_estimators,
            random_state=42,
            n_jobs=-1,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
        )

        if has_val and "split" in df2.columns:
            train_mask = base_mask & (
                df2["split"].astype(str).eq("train").to_numpy()
            )
            val_mask = base_mask & (df2["split"].astype(str).eq("val").to_numpy())
            if int(train_mask.sum()) < 2 or int(val_mask.sum()) < 1:
                raise ValueError(
                    "Validation split exists but is too small for RF sweep."
                )

            model.fit(X[train_mask], y[train_mask].astype(float))
            y_true = y[val_mask].astype(float)
            y_pred = model.predict(X[val_mask]).astype(float)
            rmse = float(np.sqrt(mean_squared_error(y_true, y_pred)))
            extra_logs["val/rmse"] = rmse
            if y_true.size >= 2:
                r2 = float(r2_score(y_true, y_pred))
                extra_logs["val/r2"] = r2
                score = r2
            else:
                score = -rmse
        else:
            y_base = y[base_mask].astype(float)
            X_base = X[base_mask]
            # Use simple 5-fold CV (non-stratified).
            n_splits = min(5, int(len(y_base)))
            n_splits = max(n_splits, 2)
            cv = KFold(n_splits=n_splits, shuffle=True, random_state=42)
            scores = cross_val_score(model, X_base, y_base, cv=cv, scoring="r2")
            score = float(np.mean(scores)) if len(scores) else 0.0
            extra_logs["cv/r2"] = score

        wandb.log({"val_score": float(score), **extra_logs})
        run.summary["val_score"] = float(score)
        run.summary["task_type"] = task_type
        run.finish()

        if float(score) > float(best_score):
            best_score = float(score)
            best_trial_number = int(trial_idx)
            best_params = {
                "n_estimators": n_estimators,
                "max_depth": max_depth,
                "min_samples_split": min_samples_split,
                "radius": int(radius),
                "n_bits": int(n_bits),
                "task_type": task_type,
            }

    wandb.agent(sweep_id, function=_trial_fn, count=int(config.n_trials))

    best_params_path = str(run_dir / f"hyperopt_best_params_wandb_rf_{task_type}.json")
    with open(best_params_path, "w") as f:
        json.dump(best_params, f, indent=2)

    return HyperoptResult(
        run_dir=run_dir,
        best_params=best_params,
        best_value=float(best_score) if best_score != -math.inf else 0.0,
        best_trial_number=int(best_trial_number),
        n_trials=int(config.n_trials),
        study_path=str(sweep_id),
        best_params_path=str(best_params_path),
        strategy="holdout-val" if has_val else "cv",
        cv_folds_used=None,
    )


def optimize_svr_wandb(
    run_dir: Path,
    *,
    config: HyperoptConfig,
    project: str,
    entity: str | None = None,
    api_key: str | None = None,
    sweep_name: str | None = None,
    radius: int = 2,
    n_bits: int = 2048,
) -> HyperoptResult:
    """Optimize SVR hyperparameters using a W&B Bayes sweep.

    Notes:
    - SVR sweep is intended for regression runs (DTA).
    - Uses holdout validation if a val split exists; otherwise uses KFold CV.
    - Logs a single scalar metric `val_score` (maximize).
    """

    wandb = _require_wandb()

    run_dir = Path(run_dir).resolve()
    dataset_path = run_dir / "dataset.csv"
    compounds_path = run_dir / "compounds.csv"
    metadata_path = run_dir / "metadata.json"

    if not dataset_path.exists() or not compounds_path.exists():
        raise FileNotFoundError(f"Expected {dataset_path} and {compounds_path}")

    df = pd.read_csv(dataset_path)
    compounds = pd.read_csv(compounds_path)

    if "molecule_chembl_id" not in df.columns or "label" not in df.columns:
        raise ValueError("dataset.csv must contain 'molecule_chembl_id' and 'label'.")
    if (
        "molecule_chembl_id" not in compounds.columns
        or "smiles" not in compounds.columns
    ):
        raise ValueError(
            "compounds.csv must contain 'molecule_chembl_id' and 'smiles'."
        )

    meta = None
    try:
        if metadata_path.exists():
            meta = json.loads(metadata_path.read_text())
    except Exception:
        meta = None

    task_type = _infer_task_type_from_metadata_or_labels(df, meta)
    if task_type != "regression":
        raise ValueError(
            "SVR sweeps are supported for regression runs only. "
            "Build a regression (DTA) dataset or choose a different model for HPO."
        )

    # Compute fingerprints once.
   
    df_comp = (
        compounds[["molecule_chembl_id", "smiles"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )
    df_feat = calculate_morgan_fingerprints(
        df_comp,
        smiles_col="smiles",
        radius=int(radius),
        n_bits=int(n_bits),
        out_col="morgan_fingerprint",
        drop_failures=True,
    )
    df_feat = (
        df_feat[["molecule_chembl_id", "morgan_fingerprint"]].dropna().drop_duplicates()
    )
    fp_map = dict(
        zip(
            df_feat["molecule_chembl_id"].astype(str),
            df_feat["morgan_fingerprint"].astype(str),
        )
    )

    df2 = df.copy()
    df2["_fp"] = df2["molecule_chembl_id"].astype(str).map(fp_map)
    df2 = df2.dropna(subset=["_fp"]).reset_index(drop=True)
    if df2.empty:
        raise ValueError("No rows left after joining fingerprints to dataset.")

    X = _bitstrings_to_numpy(df2["_fp"].astype(str).tolist(), n_bits=int(n_bits))
    y = df2["label"].astype(float).to_numpy()

    # Exclude test split for tuning when present.
    if "split" in df2.columns:
        base_mask = df2["split"].astype(str).ne("test").to_numpy()
        has_val = bool((df2.loc[base_mask, "split"].astype(str) == "val").any())
    else:
        base_mask = np.ones(len(df2), dtype=bool)
        has_val = False

    project = (project or "").strip() or "dta_gnn"
    entity = (entity or "").strip() or None
    if api_key and str(api_key).strip():
        wandb.login(key=str(api_key).strip(), relogin=True)

    # Build sweep parameter space only from enabled knobs.
    parameters: dict[str, dict] = {}

    if config.svr_optimize_C:
        parameters["C"] = {
            "distribution": "log_uniform_values",
            "min": float(config.svr_C_min),
            "max": float(config.svr_C_max),
        }
    if config.svr_optimize_epsilon:
        parameters["epsilon"] = {
            "distribution": "log_uniform_values",
            "min": float(config.svr_epsilon_min),
            "max": float(config.svr_epsilon_max),
        }

    kernel_choices = getattr(config, "svr_kernel_choices", None) or ["rbf", "linear"]
    kernel_choices = [
        str(k).strip().lower()
        for k in kernel_choices
        if str(k).strip().lower() in {"rbf", "linear"}
    ]
    if not kernel_choices:
        kernel_choices = ["rbf", "linear"]
    if config.svr_optimize_kernel:
        parameters["kernel"] = {"values": kernel_choices}

    if not parameters:
        raise ValueError(
            "No parameters selected for optimization. "
            "Enable at least one 'Optimize ...' checkbox before running a sweep."
        )

    sweep_config: dict[str, object] = {
        "name": sweep_name or "dta_gnn_svr",
        "method": "bayes",
        "metric": {"name": "val_score", "goal": "maximize"},
        "parameters": parameters,
    }

    sweep_id = wandb.sweep(sweep=sweep_config, project=project, entity=entity)

    best_score = -math.inf
    best_params: dict[str, object] = {}
    best_trial_number = -1
    trial_counter = {"i": 0}

    def _trial_fn():
        nonlocal best_score, best_params, best_trial_number

        run = wandb.init(project=project, entity=entity, config={})
        trial_idx = int(trial_counter["i"])
        trial_counter["i"] = trial_idx + 1

        sampled = dict(getattr(wandb, "config", {}) or {})



        C = float(sampled.get("C", float(getattr(config, "svr_C_default", 10.0))))
        epsilon = float(
            sampled.get("epsilon", float(getattr(config, "svr_epsilon_default", 0.1)))
        )

        k = (
            str(
                sampled.get("kernel", str(getattr(config, "svr_kernel_default", "rbf")))
            )
            .strip()
            .lower()
        )
        if k not in {"rbf", "linear"}:
            k = "rbf"

        model = SVR(kernel=k, C=C, epsilon=epsilon)

        score: float
        extra_logs: dict[str, object] = {}

        if has_val and "split" in df2.columns:
            train_mask = base_mask & (df2["split"].astype(str).eq("train").to_numpy())
            val_mask = base_mask & (df2["split"].astype(str).eq("val").to_numpy())
            if int(train_mask.sum()) < 2 or int(val_mask.sum()) < 1:
                raise ValueError(
                    "Validation split exists but is too small for SVR sweep."
                )

            model.fit(X[train_mask], y[train_mask])
            y_true = y[val_mask]
            y_pred = model.predict(X[val_mask]).astype(float)
            rmse = float(np.sqrt(mean_squared_error(y_true, y_pred)))
            extra_logs["val/rmse"] = rmse
            if y_true.size >= 2:
                r2 = float(r2_score(y_true, y_pred))
                extra_logs["val/r2"] = r2
                score = r2
            else:
                score = -rmse
        else:
            y_base = y[base_mask]
            X_base = X[base_mask]
            n_splits = min(5, int(len(y_base)))
            n_splits = max(n_splits, 2)
            cv = KFold(n_splits=n_splits, shuffle=True, random_state=42)
            scores = cross_val_score(model, X_base, y_base, cv=cv, scoring="r2")
            score = float(np.mean(scores)) if len(scores) else 0.0
            extra_logs["cv/r2"] = score
            extra_logs["cv/folds"] = int(n_splits)

        wandb.log({"val_score": float(score), **extra_logs})
        run.summary["val_score"] = float(score)
        run.summary["task_type"] = task_type
        run.finish()

        if float(score) > float(best_score):
            best_score = float(score)
            best_trial_number = int(trial_idx)
            best_params = {
                "C": float(C),
                "epsilon": float(epsilon),
                "kernel": k,
                "radius": int(radius),
                "n_bits": int(n_bits),
                "task_type": task_type,
            }

    wandb.agent(sweep_id, function=_trial_fn, count=int(config.n_trials))

    best_params_path = str(run_dir / "hyperopt_best_params_wandb_svr.json")
    with open(best_params_path, "w") as f:
        json.dump(best_params, f, indent=2)

    return HyperoptResult(
        run_dir=run_dir,
        best_params=best_params,
        best_value=float(best_score) if best_score != -math.inf else 0.0,
        best_trial_number=int(best_trial_number),
        n_trials=int(config.n_trials),
        study_path=str(sweep_id),
        best_params_path=str(best_params_path),
        strategy="holdout-val" if has_val else "cv",
        cv_folds_used=None,
    )


def optimize_gnn_wandb(
    run_dir: Path,
    *,
    config: HyperoptConfig,
    project: str,
    entity: str | None = None,
    api_key: str | None = None,
    sweep_name: str | None = None,
) -> HyperoptResult:
    """Optimize GNN hyperparameters using a W&B Bayes sweep.

    Notes:
    - Uses the existing train/val split (requires a non-empty val split).
    - Runs each trial in an isolated subdirectory so artifacts don't overwrite.
    - Logs a single scalar metric `val_score` (maximize).
    """

    wandb = _require_wandb()

    run_dir = Path(run_dir).resolve()
    dataset_path = run_dir / "dataset.csv"
    compounds_path = run_dir / "compounds.csv"
    metadata_path = run_dir / "metadata.json"

    if not dataset_path.exists() or not compounds_path.exists():
        raise FileNotFoundError(f"Expected {dataset_path} and {compounds_path}")

    df = pd.read_csv(dataset_path)
    if "split" not in df.columns or not bool((df["split"] == "val").any()):
        raise ValueError(
            "W&B sweeps currently require an explicit validation split. "
            "Rebuild the dataset with val_size > 0.0."
        )

    arch = str(getattr(config, "architecture", "gin") or "gin").strip().lower()
    if arch not in {"gin", "gcn", "gat", "sage", "pna"}:
        arch = "gin"

    project = (project or "").strip() or "dta_gnn"
    entity = (entity or "").strip() or None
    if api_key and str(api_key).strip():
        wandb.login(key=str(api_key).strip(), relogin=True)

    # Build sweep parameter space only from enabled knobs.
    parameters: dict[str, dict] = {}

    def _int_values(min_v: int, max_v: int, step: int = 1) -> list[int]:
        if step <= 1:
            return list(range(int(min_v), int(max_v) + 1))
        return list(range(int(min_v), int(max_v) + 1, int(step)))

    if config.optimize_epochs:
        parameters["epochs"] = {
            "distribution": "int_uniform",
            "min": int(config.epochs_min),
            "max": int(config.epochs_max),
        }
    if config.optimize_batch_size:
        parameters["batch_size"] = {
            "distribution": "int_uniform",
            "min": int(config.batch_size_min),
            "max": int(config.batch_size_max),
        }
    if config.optimize_lr:
        parameters["lr"] = {
            "distribution": "log_uniform_values",
            "min": float(config.lr_min),
            "max": float(config.lr_max),
        }
    if config.optimize_embedding_dim:
        parameters["embedding_dim"] = {
            "values": _int_values(
                int(config.embedding_dim_min),
                int(config.embedding_dim_max),
                step=16,
            )
        }
    if config.optimize_hidden_dim:
        parameters["hidden_dim"] = {
            "values": _int_values(
                int(config.hidden_dim_min), int(config.hidden_dim_max), step=16
            )
        }

    if config.optimize_num_layers:
        parameters["num_layers"] = {
            "distribution": "int_uniform",
            "min": int(config.num_layers_min),
            "max": int(config.num_layers_max),
        }

    if config.optimize_dropout:
        parameters["dropout"] = {
            "distribution": "uniform",
            "min": float(config.dropout_min),
            "max": float(config.dropout_max),
        }

    pooling_choices = getattr(config, "pooling_choices", None) or [
        "add",
        "mean",
        "max",
        "attention",
    ]
    pooling_choices = [
        str(x) for x in pooling_choices if str(x) in {"add", "mean", "max", "attention"}
    ]
    if not pooling_choices:
        pooling_choices = ["add", "mean", "max", "attention"]
    if config.optimize_pooling:
        parameters["pooling"] = {"values": pooling_choices}

    if config.optimize_residual:
        parameters["residual"] = {"values": [False, True]}

    if config.optimize_head_mlp_layers:
        parameters["head_mlp_layers"] = {
            "distribution": "int_uniform",
            "min": int(getattr(config, "head_mlp_layers_min", 1)),
            "max": int(getattr(config, "head_mlp_layers_max", 4)),
        }

    if arch == "gin":
        if config.optimize_gin_conv_mlp_layers:
            parameters["gin_conv_mlp_layers"] = {
                "distribution": "int_uniform",
                "min": int(getattr(config, "gin_conv_mlp_layers_min", 1)),
                "max": int(getattr(config, "gin_conv_mlp_layers_max", 4)),
            }
        if config.optimize_gin_train_eps:
            parameters["gin_train_eps"] = {"values": [False, True]}
        if config.optimize_gin_eps:
            parameters["gin_eps"] = {
                "distribution": "uniform",
                "min": float(getattr(config, "gin_eps_min", 0.0)),
                "max": float(getattr(config, "gin_eps_max", 1.0)),
            }

    if arch == "gat" and config.optimize_gat_heads:
        parameters["gat_heads"] = {
            "distribution": "int_uniform",
            "min": int(getattr(config, "gat_heads_min", 1)),
            "max": int(getattr(config, "gat_heads_max", 8)),
        }

    if arch == "sage":
        sage_aggr_choices = getattr(config, "sage_aggr_choices", None) or [
            "mean",
            "max",
            "lstm",
            "pool",
        ]
        sage_aggr_choices = [
            str(x) for x in sage_aggr_choices if str(x) in {"mean", "max", "lstm", "pool"}
        ]
        if not sage_aggr_choices:
            sage_aggr_choices = ["mean", "max", "lstm", "pool"]
        if config.optimize_sage_aggr:
            parameters["sage_aggr"] = {"values": sage_aggr_choices}

    if arch == "transformer" and config.optimize_transformer_heads:
        parameters["transformer_heads"] = {
            "distribution": "int_uniform",
            "min": int(getattr(config, "transformer_heads_min", 1)),
            "max": int(getattr(config, "transformer_heads_max", 8)),
        }

    if arch == "tag" and config.optimize_tag_k:
        parameters["tag_k"] = {
            "distribution": "int_uniform",
            "min": int(getattr(config, "tag_k_min", 1)),
            "max": int(getattr(config, "tag_k_max", 5)),
        }

    if arch == "arma":
        if config.optimize_arma_stacks:
            parameters["arma_num_stacks"] = {
                "distribution": "int_uniform",
                "min": int(getattr(config, "arma_num_stacks_min", 1)),
                "max": int(getattr(config, "arma_num_stacks_max", 3)),
            }
        if config.optimize_arma_layers:
            parameters["arma_num_layers"] = {
                "distribution": "int_uniform",
                "min": int(getattr(config, "arma_num_layers_min", 1)),
                "max": int(getattr(config, "arma_num_layers_max", 3)),
            }

    if arch == "cheb" and config.optimize_cheb_k:
        parameters["cheb_k"] = {
            "distribution": "int_uniform",
            "min": int(getattr(config, "cheb_k_min", 1)),
            "max": int(getattr(config, "cheb_k_max", 5)),
        }

    if arch == "supergat":
        if config.optimize_supergat_heads:
            parameters["supergat_heads"] = {
                "distribution": "int_uniform",
                "min": int(getattr(config, "supergat_heads_min", 1)),
                "max": int(getattr(config, "supergat_heads_max", 8)),
            }
        supergat_attention_choices = getattr(config, "supergat_attention_type_choices", None) or [
            "MX",
            "SD",
        ]
        supergat_attention_choices = [
            str(x) for x in supergat_attention_choices if str(x) in {"MX", "SD"}
        ]
        if not supergat_attention_choices:
            supergat_attention_choices = ["MX", "SD"]
        if config.optimize_supergat_attention_type:
            parameters["supergat_attention_type"] = {"values": supergat_attention_choices}

    if not parameters:
        raise ValueError(
            "No parameters selected for optimization. "
            "Enable at least one 'Optimize ...' checkbox before running a sweep."
        )

    sweep_config: dict[str, object] = {
        "name": sweep_name or f"dta_gnn_gnn_{arch}",
        "method": "bayes",
        "metric": {"name": "val_score", "goal": "maximize"},
        "parameters": parameters,
    }

    sweep_id = wandb.sweep(sweep=sweep_config, project=project, entity=entity)

    best_score = -math.inf
    best_params: dict[str, object] = {}
    best_trial_number = -1
    trial_counter = {"i": 0}

    from dta_gnn.models.gnn import GnnTrainConfig, train_gnn_on_run

    def _trial_fn():
        nonlocal best_score, best_params, best_trial_number

        run = wandb.init(project=project, entity=entity, config={})
        trial_idx = int(trial_counter["i"])
        trial_counter["i"] = trial_idx + 1
        
        print(f"\n[Trial {trial_idx}] Starting GNN training...")
        print(f"[Trial {trial_idx}] Run ID: {run.id}")

        # Pull optimized params from wandb.config; fill the rest from defaults.
        sampled = dict(getattr(wandb, "config", {}) or {})
        
        # Log all hyperparameters to wandb
        wandb.config.update({
            "trial_number": trial_idx,
            "architecture": arch,
        })

        # Architecture-specific parameters
        gin_conv_mlp_layers = int(
            sampled.get(
                "gin_conv_mlp_layers",
                int(getattr(config, "gin_conv_mlp_layers_default", 2)),
            )
        )
        gin_train_eps = bool(
            sampled.get(
                "gin_train_eps",
                bool(getattr(config, "gin_train_eps_default", False)),
            )
        )
        gin_eps = float(
            sampled.get(
                "gin_eps",
                float(getattr(config, "gin_eps_default", 0.0)),
            )
        )
        gat_heads = int(
            sampled.get(
                "gat_heads",
                int(getattr(config, "gat_heads_default", 4)),
            )
        )
        sage_aggr = str(
            sampled.get(
                "sage_aggr",
                str(getattr(config, "sage_aggr_default", "mean")),
            )
        )
        transformer_heads = int(
            sampled.get(
                "transformer_heads",
                int(getattr(config, "transformer_heads_default", 4)),
            )
        )
        tag_k = int(
            sampled.get(
                "tag_k",
                int(getattr(config, "tag_k_default", 2)),
            )
        )
        arma_num_stacks = int(
            sampled.get(
                "arma_num_stacks",
                int(getattr(config, "arma_num_stacks_default", 1)),
            )
        )
        arma_num_layers = int(
            sampled.get(
                "arma_num_layers",
                int(getattr(config, "arma_num_layers_default", 1)),
            )
        )
        cheb_k = int(
            sampled.get(
                "cheb_k",
                int(getattr(config, "cheb_k_default", 2)),
            )
        )
        supergat_heads = int(
            sampled.get(
                "supergat_heads",
                int(getattr(config, "supergat_heads_default", 4)),
            )
        )
        supergat_attention_type = str(
            sampled.get(
                "supergat_attention_type",
                str(getattr(config, "supergat_attention_type_default", "MX")),
            )
        )

        cfg = GnnTrainConfig(
            architecture=arch,
            epochs=int(sampled.get("epochs", int(getattr(config, "epochs_default", 20)))),
            batch_size=int(sampled.get("batch_size", int(getattr(config, "batch_size_default", 64)))),
            lr=float(sampled.get("lr", 1e-3)),
            embedding_dim=int(sampled.get("embedding_dim", int(getattr(config, "embedding_dim_default", 128)))),
            device=getattr(config, "device", None),
            hidden_dim=int(sampled.get("hidden_dim", int(getattr(config, "hidden_dim_default", 128)))),
            dropout=float(
                sampled.get(
                    "dropout", float(getattr(config, "dropout_default", 0.1))
                )
            ),
            pooling=str(
                sampled.get(
                    "pooling", str(getattr(config, "pooling_default", "add"))
                )
            ),
            residual=bool(
                sampled.get(
                    "residual", bool(getattr(config, "residual_default", False))
                )
            ),
            head_mlp_layers=int(
                sampled.get(
                    "head_mlp_layers",
                    int(getattr(config, "head_mlp_layers_default", 2)),
                )
            ),
            gin_conv_mlp_layers=gin_conv_mlp_layers,
            gin_train_eps=gin_train_eps,
            gin_eps=gin_eps,
            gat_heads=gat_heads,
            sage_aggr=sage_aggr,
            transformer_heads=transformer_heads,
            tag_k=tag_k,
            arma_num_stacks=arma_num_stacks,
            arma_num_layers=arma_num_layers,
            cheb_k=cheb_k,
            supergat_heads=supergat_heads,
            supergat_attention_type=supergat_attention_type,
            num_layers=int(
                sampled.get(
                    "num_layers",
                    int(getattr(config, "num_layers_default", 3)),
                )
            ),
        )

        trial_dir = run_dir / f"_wandb_{arch}_{trial_idx:03d}_{run.id}"
        trial_dir.mkdir(exist_ok=True)
        shutil.copy2(dataset_path, trial_dir / "dataset.csv")
        shutil.copy2(compounds_path, trial_dir / "compounds.csv")
        if metadata_path.exists():
            shutil.copy2(metadata_path, trial_dir / "metadata.json")

        print(f"[Trial {trial_idx}] Training GNN with config: epochs={cfg.epochs}, batch_size={cfg.batch_size}, lr={cfg.lr:.6f}, embedding_dim={cfg.embedding_dim}")
        res = train_gnn_on_run(trial_dir, config=cfg, wandb_run=run)
        score = _score_from_gnn_metrics(res.task_type, res.metrics)
        print(f"[Trial {trial_idx}] Training complete. Score: {score:.4f}")

        # Log all metrics from all splits
        all_metrics = {}
        splits_metrics = (res.metrics or {}).get("splits", {}) or {}
        
        # Log validation metrics
        val_metrics = splits_metrics.get("val", {}) or {}
        all_metrics.update({
            "val_score": float(score),
            "val/roc_auc": val_metrics.get("roc_auc"),
            "val/accuracy": val_metrics.get("accuracy"),
            "val/rmse": val_metrics.get("rmse"),
            "val/mae": val_metrics.get("mae"),
            "val/r2": val_metrics.get("r2"),
        })
        
        # Log training metrics
        train_metrics = splits_metrics.get("train", {}) or {}
        all_metrics.update({
            "train/roc_auc": train_metrics.get("roc_auc"),
            "train/accuracy": train_metrics.get("accuracy"),
            "train/rmse": train_metrics.get("rmse"),
            "train/mae": train_metrics.get("mae"),
            "train/r2": train_metrics.get("r2"),
        })
        
        # Log test metrics if available
        test_metrics = splits_metrics.get("test", {}) or {}
        if test_metrics:
            all_metrics.update({
                "test/roc_auc": test_metrics.get("roc_auc"),
                "test/accuracy": test_metrics.get("accuracy"),
                "test/rmse": test_metrics.get("rmse"),
                "test/mae": test_metrics.get("mae"),
                "test/r2": test_metrics.get("r2"),
            })
        
        # Remove None values before logging
        all_metrics = {k: v for k, v in all_metrics.items() if v is not None}
        wandb.log(all_metrics)

        run.summary["val_score"] = float(score)
        run.summary["run_dir"] = str(trial_dir)
        run.summary["architecture"] = arch
        run.summary["task_type"] = res.task_type
        run.summary["trial_number"] = trial_idx
        run.summary["is_best"] = float(score) > float(best_score)
        run.finish()

        if float(score) > float(best_score):
            best_score = float(score)
            best_trial_number = int(trial_idx)
            # Return the *sampled* params plus fixed architecture for reproducibility.
            best_params = {"architecture": arch, **{k: v for k, v in sampled.items()}}
            print(f"[Trial {trial_idx}] New best score: {best_score:.4f} (trial #{best_trial_number})")
        else:
            print(f"[Trial {trial_idx}] Score: {score:.4f} (best so far: {best_score:.4f})")

    print(f"\n{'='*60}")
    print(f"Starting W&B sweep with {config.n_trials} trials")
    print(f"Sweep ID: {sweep_id}")
    print(f"Architecture: {arch}")
    print(f"{'='*60}\n")
    
    wandb.agent(sweep_id, function=_trial_fn, count=int(config.n_trials))

    best_params_path = str(run_dir / f"hyperopt_best_params_wandb_{arch}.json")
    with open(best_params_path, "w") as f:
        json.dump(best_params, f, indent=2)

    print(f"\n{'='*60}")
    print(f"Sweep completed!")
    print(f"Best score: {best_score:.4f} (trial #{best_trial_number})")
    print(f"Best params saved to: {best_params_path}")
    print(f"{'='*60}\n")

    return HyperoptResult(
        run_dir=run_dir,
        best_params=best_params,
        best_value=float(best_score) if best_score != -math.inf else 0.0,
        best_trial_number=int(best_trial_number),
        n_trials=int(config.n_trials),
        study_path=str(sweep_id),
        best_params_path=str(best_params_path),
        strategy="holdout-val",
        cv_folds_used=None,
    )


# Backwards-compatible aliases: redirect to W&B Bayes versions
# All optimization now uses W&B Bayes sweeps instead of Optuna
optimize_random_forest = optimize_random_forest_wandb
optimize_gnn = optimize_gnn_wandb
optimize_gin = optimize_gnn_wandb  # Historical alias



"""Tests for I/O sources (web and SQLite)."""

import pytest
from unittest.mock import MagicMock, patch

from dta_gnn.io.web_source import ChemblWebSource
from dta_gnn.io.sqlite_source import ChemblSQLiteSource


def test_web_source_activities():
    with patch("dta_gnn.io.web_source.new_client") as mock_client:
        # Setup mock return
        mock_activity = MagicMock()
        mock_client.activity = mock_activity

        # Mocking the filter chain is tricky, simplified assumption
        mock_query = MagicMock()
        mock_activity.filter.return_value = mock_query
        mock_query.filter.return_value = mock_query

        # Iteration over query
        mock_query.__iter__.return_value = iter(
            [
                {
                    "molecule_chembl_id": "CHEMBL1",
                    "target_chembl_id": "CHEMBL2",
                    "standard_type": "IC50",
                    "standard_value": "100",
                    "standard_units": "nM",
                    "standard_relation": "=",
                    "pchembl_value": "7.0",
                }
            ]
        )

        source = ChemblWebSource()
        df = source.fetch_activities(target_ids=["CHEMBL2"])

        assert len(df) == 1
        assert df.iloc[0]["molecule_chembl_id"] == "CHEMBL1"
        assert df.iloc[0]["standard_value"] == "100"


def test_sqlite_source_activities(tmp_path):
    # Create a dummy sqlite db
    db_file = tmp_path / "chembl.db"
    import sqlite3

    conn = sqlite3.connect(db_file)
    cursor = conn.cursor()

    cursor.execute(
        "CREATE TABLE molecule_dictionary (molregno INTEGER, chembl_id TEXT)"
    )
    cursor.execute(
        "CREATE TABLE target_dictionary (tid INTEGER, chembl_id TEXT, organism TEXT)"
    )
    cursor.execute("CREATE TABLE assays (assay_id INTEGER, tid INTEGER)")
    cursor.execute("CREATE TABLE docs (doc_id INTEGER, year INTEGER)")
    cursor.execute(
        "CREATE TABLE activities (molregno INTEGER, assay_id INTEGER, doc_id INTEGER, standard_type TEXT, standard_value REAL, standard_units TEXT, standard_relation TEXT, pchembl_value REAL)"
    )

    cursor.execute("INSERT INTO molecule_dictionary VALUES (1, 'CHEMBL1')")
    cursor.execute("INSERT INTO target_dictionary VALUES (1, 'CHEMBL2', 'Human')")
    cursor.execute("INSERT INTO assays VALUES (1, 1)")
    cursor.execute("INSERT INTO docs VALUES (1, 2020)")
    cursor.execute(
        "INSERT INTO activities VALUES (1, 1, 1, 'IC50', 100, 'nM', '=', 7.0)"
    )

    conn.commit()
    conn.close()

    source = ChemblSQLiteSource(str(db_file))
    df = source.fetch_activities(target_ids=["CHEMBL2"])

    assert len(df) == 1
    assert df.iloc[0]["molecule_chembl_id"] == "CHEMBL1"
    assert df.iloc[0]["standard_value"] == 100.0  # SQLite stores as REAL


class TestChemblWebSourceAdvanced:
    """Additional tests for ChEMBL web source."""

    def test_fetch_activities_empty_result(self):
        """Test handling of empty results from API."""
        with patch("dta_gnn.io.web_source.new_client") as mock_client:
            mock_activity = MagicMock()
            mock_client.activity = mock_activity
            mock_query = MagicMock()
            mock_activity.filter.return_value = mock_query
            mock_query.filter.return_value = mock_query
            mock_query.__iter__.return_value = iter([])

            source = ChemblWebSource()
            df = source.fetch_activities(target_ids=["NONEXISTENT"])

            assert len(df) == 0


class TestChemblSQLiteSourceAdvanced:
    """Additional tests for SQLite source."""

    def test_multiple_targets(self, tmp_path):
        """Test fetching activities for multiple targets."""
        db_file = tmp_path / "test.db"
        import sqlite3

        conn = sqlite3.connect(db_file)
        cursor = conn.cursor()

        cursor.execute(
            "CREATE TABLE molecule_dictionary (molregno INTEGER, chembl_id TEXT)"
        )
        cursor.execute(
            "CREATE TABLE target_dictionary (tid INTEGER, chembl_id TEXT, organism TEXT)"
        )
        cursor.execute("CREATE TABLE assays (assay_id INTEGER, tid INTEGER)")
        cursor.execute("CREATE TABLE docs (doc_id INTEGER, year INTEGER)")
        cursor.execute(
            "CREATE TABLE activities (molregno INTEGER, assay_id INTEGER, doc_id INTEGER, standard_type TEXT, standard_value REAL, standard_units TEXT, standard_relation TEXT, pchembl_value REAL)"
        )

        cursor.execute("INSERT INTO molecule_dictionary VALUES (1, 'CHEMBL1')")
        cursor.execute("INSERT INTO molecule_dictionary VALUES (2, 'CHEMBL2')")
        cursor.execute("INSERT INTO target_dictionary VALUES (1, 'T1', 'Human')")
        cursor.execute("INSERT INTO target_dictionary VALUES (2, 'T2', 'Human')")
        cursor.execute("INSERT INTO assays VALUES (1, 1)")
        cursor.execute("INSERT INTO assays VALUES (2, 2)")
        cursor.execute("INSERT INTO docs VALUES (1, 2020)")
        cursor.execute(
            "INSERT INTO activities VALUES (1, 1, 1, 'IC50', 100, 'nM', '=', 7.0)"
        )
        cursor.execute(
            "INSERT INTO activities VALUES (2, 2, 1, 'Ki', 50, 'nM', '=', 7.3)"
        )

        conn.commit()
        conn.close()

        source = ChemblSQLiteSource(str(db_file))
        df = source.fetch_activities(target_ids=["T1", "T2"])

        assert len(df) == 2

    def test_nonexistent_db(self, tmp_path):
        """Test handling of nonexistent database."""
        with pytest.raises(Exception):
            ChemblSQLiteSource(str(tmp_path / "nonexistent.db"))



import sqlite3
import pandas as pd
from typing import Optional, List
from loguru import logger
from dta_gnn.io.chembl_source import ChemblSource


class ChemblSQLiteSource(ChemblSource):
    """ChEMBL data source using a local SQLite database dump."""

    def __init__(self, db_path: str):
        import os

        if not os.path.exists(db_path):
            raise FileNotFoundError(f"ChEMBL SQLite file not found at: {db_path}")
        self.db_path = db_path

    def _get_conn(self):
        # connect in RO mode to prevent accidental creation if path is wrong (though check above handles most)
        # But sqlite3.connect can still create if race condition or other issue?
        # Standard connect is fine if we validated path.
        uri = f"file:{self.db_path}?mode=ro"
        conn = sqlite3.connect(uri, uri=True)

        # Validate schema
        try:
            cursor = conn.cursor()
            cursor.execute(
                "SELECT name FROM sqlite_master WHERE type='table' AND name='activities';"
            )
            if not cursor.fetchone():
                conn.close()
                raise ValueError(
                    "Database does not contain 'activities' table. Is this a valid ChEMBL DB?"
                )
        except Exception:
            conn.close()
            raise

        return conn

    def fetch_activities(
        self,
        target_ids: Optional[List[str]] = None,
        molecule_ids: Optional[List[str]] = None,
        standard_types: Optional[List[str]] = None,
        progress_callback: Optional[callable] = None,
    ) -> pd.DataFrame:        # Optimize query by starting from target_dictionary when filtering by target_ids
        # This allows early filtering and better query plan
        if target_ids:
            query = """
                SELECT 
                    md.chembl_id as molecule_chembl_id,
                    td.chembl_id as target_chembl_id,
                    act.standard_type,
                    act.standard_value,
                    act.standard_units,
                    act.standard_relation,
                    act.pchembl_value,
                    d.year
                FROM target_dictionary td
                JOIN assays ass ON td.tid = ass.tid
                JOIN activities act ON ass.assay_id = act.assay_id
                JOIN molecule_dictionary md ON act.molregno = md.molregno
                LEFT JOIN docs d ON act.doc_id = d.doc_id
                WHERE td.chembl_id IN ({})
            """.format(",".join("?" * len(target_ids)))
            params = list(target_ids)
        else:
            query = """
                SELECT 
                    md.chembl_id as molecule_chembl_id,
                    td.chembl_id as target_chembl_id,
                    act.standard_type,
                    act.standard_value,
                    act.standard_units,
                    act.standard_relation,
                    act.pchembl_value,
                    d.year
                FROM activities act
                JOIN molecule_dictionary md ON act.molregno = md.molregno
                JOIN assays ass ON act.assay_id = ass.assay_id
                JOIN target_dictionary td ON ass.tid = td.tid
                LEFT JOIN docs d ON act.doc_id = d.doc_id
                WHERE 1=1
            """
            params = []

        if molecule_ids:
            placeholders = ",".join("?" * len(molecule_ids))
            if target_ids:
                query += f" AND md.chembl_id IN ({placeholders})"
            else:
                query += f" AND md.chembl_id IN ({placeholders})"
            params.extend(molecule_ids)

        if standard_types:
            placeholders = ",".join("?" * len(standard_types))
            query += f" AND act.standard_type IN ({placeholders})"
            params.extend(standard_types)

        logger.info("Executing SQL query to fetch activities (this may take a while for large databases)...")
        logger.debug(f"Query filters: target_ids={target_ids}, molecule_ids={molecule_ids}, standard_types={standard_types}")
        
        with self._get_conn() as conn:
            # Set a longer timeout for large queries
            conn.execute("PRAGMA busy_timeout = 30000")  # 30 seconds
            df = pd.read_sql_query(query, conn, params=params)
            logger.info(f"Query completed. Retrieved {len(df)} activity records.")
            return df

    def fetch_molecules(self, molecule_ids: List[str]) -> pd.DataFrame:
        chunk_size = 900
        dfs = []
        
        logger.info(f"Fetching molecules for {len(molecule_ids)} molecule IDs (in chunks of {chunk_size})...")

        for i in range(0, len(molecule_ids), chunk_size):
            chunk = molecule_ids[i : i + chunk_size]
            query = """
                SELECT 
                    md.chembl_id as molecule_chembl_id,
                    cs.canonical_smiles as smiles
                FROM molecule_dictionary md
                JOIN compound_structures cs ON md.molregno = cs.molregno
                WHERE md.chembl_id IN ({})
            """.format(
                ",".join("?" * len(chunk))
            )

            with self._get_conn() as conn:
                dfs.append(pd.read_sql_query(query, conn, params=chunk))

        if not dfs:
            return pd.DataFrame(columns=["molecule_chembl_id", "smiles"])

        result = pd.concat(dfs, ignore_index=True)
        logger.info(f"Fetched {len(result)} molecule records.")
        return result

    def fetch_targets(self, target_ids: List[str]) -> pd.DataFrame:
        # Note: This join is simplified. Targets (tid) link to components which have sequences.
        # target_dictionary -(target_components)-> component_sequences
        query = """
            SELECT 
                td.chembl_id as target_chembl_id,
                cseq.sequence,
                td.organism
            FROM target_dictionary td
            JOIN target_components tc ON td.tid = tc.tid
            JOIN component_sequences cseq ON tc.component_id = cseq.component_id
            WHERE td.chembl_id IN ({})
        """.format(
            ",".join("?" * len(target_ids))
        )

        with self._get_conn() as conn:
            return pd.read_sql_query(query, conn, params=target_ids)



import os
import requests
import tarfile
from tqdm import tqdm
from pathlib import Path
from loguru import logger

LATEST_CHEMBL_VERSION = "36"  # Manually pinning for now to ensure stability
BASE_URL = (
    "https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/"
    "chembl_{}/chembl_{}_sqlite.tar.gz"
)


def download_chembl_db(
    version: str = LATEST_CHEMBL_VERSION, output_dir: str = "."
) -> str:
    """
    Download and extract ChEMBL SQLite database.
    Returns path to the extracted .db file.
    """
    version = str(version)
    url = BASE_URL.format(version, version)
    filename = f"chembl_{version}_sqlite.tar.gz"
    output_path = Path(output_dir) / filename

    logger.info(f"Downloading ChEMBL {version} from {url}...")

    # Stream download with progress
    if not output_path.exists():
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                total_size = int(r.headers.get("content-length", 0))

                with open(output_path, "wb") as f, tqdm(
                    desc=filename,
                    total=total_size,
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar:
                    for chunk in r.iter_content(chunk_size=8192):
                        size = f.write(chunk)
                        bar.update(size)
        except Exception as e:
            if output_path.exists():
                os.remove(output_path)
            raise RuntimeError(f"Download failed: {e}")
    else:
        logger.info("Archive already exists, skipping download.")

    # Extract
    logger.info("Extracting...")
    extract_path = Path(output_dir)

    # The tar usually contains a folder like chembl_33/chembl_33_sqlite/chembl_33.db
    # We strip components? Tarfile doesn't support strip_components easily.
    # We extract and then find the db file.

    # Check if DB already extracted? (Simplified check)
    # The internal structure is typically: chembl_33/chembl_33_sqlite/chembl_33.db
    # Let's extract everything and find it.

    with tarfile.open(output_path, "r:gz") as tar:
        # Security check for tarbomb/absolute paths?
        # Python 3.12 has filter='data', but let's just do default as we trust EBI.
        tar.extractall(path=output_dir)

        # Locate the .db file
        for member in tar.getmembers():
            if member.name.endswith(".db"):
                logger.info(f"Found DB at {member.name}")
                # We return this path.
                return str(extract_path / member.name)

    raise FileNotFoundError("Could not find .db file in archive")



"""Tests for molecule graph construction."""

import numpy as np

from dta_gnn.features.molecule_graphs import build_graphs_2d, smiles_to_graph_2d


def test_smiles_to_graph_2d_shapes():
    g = smiles_to_graph_2d(molecule_chembl_id="CHEMBL_TEST", smiles="CCO")
    assert g is not None

    assert g.atom_type.shape == (3,)
    assert g.atom_type.dtype == np.int64

    assert g.atom_feat.shape == (3, 6)
    assert g.atom_feat.dtype == np.float32

    # C-C and C-O bonds => 2 bonds => 4 directed edges
    assert g.edge_index.shape == (2, 4)
    assert g.edge_index.dtype == np.int64

    assert g.edge_attr.shape == (4, 6)
    assert g.edge_attr.dtype == np.float32


def test_build_graphs_2d_drops_invalid_smiles_by_default():
    graphs = build_graphs_2d(
        molecules=[("A", "CCO"), ("B", "not_a_smiles")],
        drop_failures=True,
    )
    assert len(graphs) == 1
    assert graphs[0].molecule_chembl_id == "A"


class TestSmilesToGraph2dAdvanced:
    """Additional tests for SMILES to graph conversion."""

    def test_benzene_ring(self):
        """Test graph construction for benzene."""
        g = smiles_to_graph_2d(molecule_chembl_id="benzene", smiles="c1ccccc1")

        assert g is not None
        assert g.atom_type.shape == (6,)
        # Benzene has 6 bonds => 12 directed edges
        assert g.edge_index.shape[1] == 12

    def test_molecule_with_heteroatoms(self):
        """Test graph construction for molecules with heteroatoms."""
        g = smiles_to_graph_2d(molecule_chembl_id="test", smiles="CCN")

        assert g is not None
        assert len(g.atom_type) == 3

    def test_complex_molecule(self):
        """Test graph construction for complex molecule (caffeine)."""
        caffeine = "Cn1cnc2c1c(=O)n(c(=O)n2C)C"
        g = smiles_to_graph_2d(molecule_chembl_id="caffeine", smiles=caffeine)

        assert g is not None
        assert len(g.atom_type) > 0

    def test_invalid_smiles_returns_none(self):
        """Test that invalid SMILES returns None."""
        g = smiles_to_graph_2d(molecule_chembl_id="invalid", smiles="not_valid")

        assert g is None

    def test_edge_features_dimensions(self):
        """Test edge feature dimensions."""
        g = smiles_to_graph_2d(molecule_chembl_id="test", smiles="CCO")

        assert g.edge_attr.shape[1] == 6
        assert g.edge_attr.dtype == np.float32


class TestBuildGraphs2dAdvanced:
    """Additional tests for batch graph construction."""

    def test_all_valid(self):
        """Test with all valid SMILES."""
        molecules = [("M1", "CCO"), ("M2", "c1ccccc1"), ("M3", "CC(=O)O")]

        graphs = build_graphs_2d(molecules=molecules)

        assert len(graphs) == 3

    def test_all_invalid(self):
        """Test with all invalid SMILES."""
        molecules = [("M1", "invalid1"), ("M2", "invalid2")]

        graphs = build_graphs_2d(molecules=molecules, drop_failures=True)

        assert len(graphs) == 0

    def test_empty_list(self):
        """Test with empty molecule list."""
        graphs = build_graphs_2d(molecules=[])

        assert len(graphs) == 0

    def test_graph_ids_preserved(self):
        """Test that molecule IDs are preserved."""
        molecules = [("CHEMBL123", "CCO"), ("CHEMBL456", "CCC")]

        graphs = build_graphs_2d(molecules=molecules)

        ids = [g.molecule_chembl_id for g in graphs]
        assert "CHEMBL123" in ids
        assert "CHEMBL456" in ids




import pytest
import sys
from unittest.mock import MagicMock, patch

import pandas as pd
import numpy as np

from dta_gnn.splits.strategies import split_random, split_cold_drug_scaffold, split_temporal

def test_split_random():
    """Test standard random splitting."""
    df = pd.DataFrame({"smiles": ["C"] * 100, "y": range(100)})
    result, train, val, test = split_random(df, test_size=0.2, val_size=0.1, seed=42)
    
    assert len(train) == 70
    assert len(val) == 10
    assert len(test) == 20
    assert len(result) == 100
    assert "split" in result.columns
    assert set(result["split"].unique()) == {"train", "val", "test"}
    
    # Test with no validation
    result, train, val, test = split_random(df, test_size=0.2, val_size=0.0)
    assert len(train) == 80
    assert len(val) == 0
    assert len(test) == 20

def test_split_random_empty():
    """Test random splitting with empty input."""
    df = pd.DataFrame(columns=["smiles", "y"])
    result, _, _, _ = split_random(df)
    assert len(result) == 0
    assert list(result.columns) == ["smiles", "y"]

def test_split_cold_drug_scaffold_basic():
    """Test scaffold splitting basics."""
    # Create molecules with distinct scaffolds
    # Benzene vs Cyclohexane vs something else
    smiles = ["c1ccccc1", "C1CCCCC1", "CCO", "CCN", "CCC"]
    df = pd.DataFrame({"smiles": smiles, "id": range(5)})
    
    # We can mock MurckoScaffold to force specific scaffold groupings if we want to be robust against RDKit behavior
    # mapping: smi -> scaffold
    scaffold_map = {
        "c1ccccc1": "S1",
        "C1CCCCC1": "S2",
        "CCO": "S3",
        "CCN": "S3", # Same scaffold
        "CCC": "S4"  # Distinct
    }
    
    with patch("dta_gnn.splits.strategies.MurckoScaffold.MurckoScaffoldSmiles", side_effect=lambda x: scaffold_map.get(x, "")):
        # 5 mols. S1(1), S2(1), S3(2), S4(1).
        # Test size 0.2 (1 mol). Val size 0.2 (1 mol).
        # Should likely put S3 in train?
        df_split = split_cold_drug_scaffold(df, test_size=0.2, val_size=0.2, seed=42)
        
        assert "split" in df_split.columns
        counts = df_split["split"].value_counts()
        # With small N, exact counts depend on shuffle and greedy allocation
        assert "train" in counts.index
        assert "test" in counts.index

def test_split_cold_drug_scaffold_invalid():
    """Test scaffold splitting with invalid SMILES triggering exception handler."""
    df = pd.DataFrame({"smiles": ["INVALID_SMILES", "C"], "id": [1, 2]})
    
    # Mock Murcko to raise Exception for one
    def side_effect(x):
        if x == "INVALID_SMILES":
            raise ValueError("Bad SMILES")
        return "C"
        
    with patch("dta_gnn.splits.strategies.MurckoScaffold.MurckoScaffoldSmiles", side_effect=side_effect):
        df_split = split_cold_drug_scaffold(df)
        # Invalid SMILES gets skipped in scaffold dict?
        # Function catches Exception (line 63) and does `pass`.
        # So it's not added to `scaffolds`.
        # So it's not in `scaffold_sets`.
        # So it's not assigned a split?
        # Code: `df.loc[idxs, "split"] = ...`.
        # The invalid row will have NaN split.
        
        assert "split" in df_split.columns
        # One valid row gets split (likely train)
        # Invalid row remains untouched (NaN split) or whatever default was
        # The function initializes split column? No.
        # So invalid row will have NaN in "split".
        invalid_row = df_split[df_split["smiles"] == "INVALID_SMILES"]
        if "split" in invalid_row.columns:
             assert pd.isna(invalid_row.iloc[0]["split"])

def test_split_temporal():
    """Test temporal splitting."""
    df = pd.DataFrame({
        "smiles": ["C"]*10,
        "year": [2000, 2010, 2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027]
    })
    
    # Split year 2022. 
    # < 2022: 2000, 2010, 2020, 2021 (4 items) -> Train (split into Train/Val)
    # >= 2022: 6 items -> Test
    
    result = split_temporal(df, year_col="year", split_year=2022, val_size=0.0)
    
    train = result[result["split"] == "train"]
    test = result[result["split"] == "test"]
    
    assert len(train) == 4
    assert len(test) == 6
    assert all(train["year"] < 2022)
    assert all(test["year"] >= 2022)

def test_split_temporal_with_val():
    """Test temporal splitting with validation set."""
    df = pd.DataFrame({
        "smiles": ["C"]*100,
        "year": [2000]*50 + [2025]*50
    })
    
    # Val size 0.1 of remaining train? Code: `train_test_split(train_df, test_size=val_size)`
    # Inputs: test_size is not param to split_temporal! Only val_size.
    # split_year determines test set.
    
    result = split_temporal(df, split_year=2022, val_size=0.1)
    
    test_len = len(result[result["split"] == "test"]) # 50
    train_val_len = 50
    
    val_len = len(result[result["split"] == "val"])
    train_len = len(result[result["split"] == "train"])
    
    assert test_len == 50
    assert val_len == 5 # 0.1 * 50
    assert train_len == 45

def test_split_temporal_edge_cases():
    """Test temporal splitting edge cases (bad year, empty result)."""
    df = pd.DataFrame({
        "smiles": ["C"],
        "year": ["not_a_year"]
    })
    
    # Should coerce to NaN. NaN < 2022 is False?
    result = split_temporal(df, split_year=2022)
    # NaN >= 2022 False.
    # Both masks False?
    # Then result is empty?
    # Code: `dfs_to_concat = [d for d in [train, val, test] if len(d) > 0]`.
    # If all empty, returns valid empty DF with cols.
    assert len(result) == 0



from __future__ import annotations

import re
import sqlite3
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Mapping


_UNIPROT_RE = re.compile(
    r"^[A-NR-Z][0-9][A-Z0-9]{3}[0-9]$|^[OPQ][0-9][A-Z0-9]{3}[0-9]$"
)
_CHEMBL_TARGET_RE = re.compile(r"^CHEMBL[0-9]+$")


@dataclass(frozen=True)
class UniProtToChEMBLResult:
    resolved_target_chembl_ids: list[str]
    per_input: Mapping[str, list[str]]
    unmapped: list[str]


def parse_uniprot_accessions(text: str) -> list[str]:
    if not (text or "").strip():
        raise ValueError("No UniProt accessions provided")

    raw = re.split(r"[\s,;]+", text.strip())
    accessions = [t.upper() for t in raw if t]

    bad = [a for a in accessions if not _UNIPROT_RE.match(a)]
    if bad:
        raise ValueError(f"Invalid UniProt accession(s): {', '.join(bad)}")

    # preserve order, unique
    seen: set[str] = set()
    out: list[str] = []
    for a in accessions:
        if a not in seen:
            out.append(a)
            seen.add(a)
    return out


def parse_chembl_target_ids(text: str) -> list[str]:
    if not (text or "").strip():
        raise ValueError("No ChEMBL target IDs provided")

    raw = re.split(r"[\s,;]+", text.strip())
    ids = [t.upper() for t in raw if t]

    bad = [t for t in ids if not _CHEMBL_TARGET_RE.match(t)]
    if bad:
        raise ValueError(f"Invalid ChEMBL target ID(s): {', '.join(bad)}")

    seen: set[str] = set()
    out: list[str] = []
    for t in ids:
        if t not in seen:
            out.append(t)
            seen.add(t)
    return out


def map_uniprot_to_chembl_targets_sqlite(
    sqlite_path: str | Path,
    accessions: Iterable[str],
) -> UniProtToChEMBLResult:
    path = Path(sqlite_path)
    if not path.exists():
        raise FileNotFoundError(f"ChEMBL SQLite DB not found: {path}")

    input_accessions = [a.upper() for a in accessions]
    per_input: dict[str, list[str]] = {a: [] for a in input_accessions}

    if not input_accessions:
        return UniProtToChEMBLResult(
            resolved_target_chembl_ids=[], per_input=per_input, unmapped=[]
        )

    placeholders = ",".join(["?"] * len(input_accessions))
    query = f"""
        SELECT cs.accession, td.chembl_id
        FROM component_sequences cs
        JOIN target_components tc ON tc.component_id = cs.component_id
        JOIN target_dictionary td ON td.tid = tc.tid
        WHERE cs.accession IN ({placeholders})
    """

    with sqlite3.connect(str(path)) as conn:
        rows = conn.execute(query, input_accessions).fetchall()

    for accession, chembl_id in rows:
        if accession is None or chembl_id is None:
            continue
        a = str(accession).upper()
        t = str(chembl_id).upper()
        if a in per_input and t not in per_input[a]:
            per_input[a].append(t)

    resolved = sorted({tid for tids in per_input.values() for tid in tids})
    unmapped = [a for a, tids in per_input.items() if not tids]

    return UniProtToChEMBLResult(
        resolved_target_chembl_ids=resolved,
        per_input=per_input,
        unmapped=unmapped,
    )


def map_uniprot_to_chembl_targets_web(
    accessions: Iterable[str],
) -> UniProtToChEMBLResult:
    """Web-based UniProt→ChEMBL mapping.

    This is implemented as a thin fallback to the existing ChEMBL web client
    logic in the app. It keeps the API stable for the UI.
    """

    from dta_gnn.io.web_source import ChemblWebSource

    input_accessions = [a.upper() for a in accessions]
    per_input: dict[str, list[str]] = {a: [] for a in input_accessions}

    src = ChemblWebSource()
    for a in input_accessions:
        try:
            # Best-effort: match targets whose component accession matches.
            targets = src.get_targets(accession=a)
            per_input[a] = sorted(
                {t["target_chembl_id"] for t in targets if "target_chembl_id" in t}
            )
        except Exception:
            per_input[a] = []

    resolved = sorted({tid for tids in per_input.values() for tid in tids})
    unmapped = [a for a, tids in per_input.items() if not tids]

    return UniProtToChEMBLResult(
        resolved_target_chembl_ids=resolved,
        per_input=per_input,
        unmapped=unmapped,
    )



from __future__ import annotations

import pandas as pd
from loguru import logger


def calculate_morgan_fingerprints(
    df: pd.DataFrame,
    smiles_col: str = "smiles",
    radius: int = 2,
    n_bits: int = 2048,
    *,
    out_col: str = "morgan_fingerprint",
    drop_failures: bool = True,
) -> pd.DataFrame:
    """Calculate Morgan fingerprints for molecules in the DataFrame.

    Returns a copy of `df` with an added fingerprint column containing bitstrings.

    Args:
        df: Input dataframe
        smiles_col: Column containing SMILES strings
        radius: Morgan radius (2 => ECFP4)
        n_bits: Fingerprint bit length
        out_col: Output column name
        drop_failures: Whether to drop rows that fail featurization
    """

    from rdkit import Chem
    from rdkit.Chem import AllChem

    logger.info(f"Calculating Morgan fingerprints (r={radius}, n={n_bits})...")

    fps: list[str | None] = []
    indices_to_drop: list[int] = []

    # Prefer the new generator API if available.
    mfgen = None
    try:
        mfgen = AllChem.GetMorganGenerator(radius=int(radius), fpSize=int(n_bits))
    except Exception:
        mfgen = None

    for idx, row in df.iterrows():
        smi = row.get(smiles_col)
        if smi is None or pd.isna(smi) or not str(smi).strip():
            fps.append(None)
            indices_to_drop.append(int(idx))
            continue

        mol = Chem.MolFromSmiles(str(smi))
        if mol is None:
            fps.append(None)
            indices_to_drop.append(int(idx))
            continue

        try:
            if mfgen is not None:
                fp = mfgen.GetFingerprint(mol)
                fps.append(fp.ToBitString())
            else:
                fp = AllChem.GetMorganFingerprintAsBitVect(
                    mol, int(radius), nBits=int(n_bits)
                )
                fps.append(fp.ToBitString())
        except Exception:
            fps.append(None)
            indices_to_drop.append(int(idx))

    out = df.copy()
    out[out_col] = fps

    if drop_failures and indices_to_drop:
        logger.warning(f"Failed to featurize {len(indices_to_drop)} molecules.")
        out = out.drop(indices_to_drop)

    return out


__all__ = [
    "calculate_morgan_fingerprints",
]



"""Artifact collection and export utilities for run directories."""

from __future__ import annotations

import os
import zipfile
from pathlib import Path

import pandas as pd


def artifact_keys_in_zip() -> list[str]:
    """Stable list of artifact keys included in artifacts.zip.

    Keep this centralized so handlers don't duplicate long lists.
    """
    return [
        "dataset",
        "targets",
        "compounds",
        "metadata",
        "molecule_features",
        "protein_features",
        "model",
        "model_metrics",
        "model_predictions",
        "model_gnn",
        "model_metrics_gnn",
        "model_predictions_gnn",
        "encoder_gnn",
        "encoder_gnn_config",
        "molecule_embeddings",
    ]


def collect_artifacts(
    *,
    run_dir: str | None,
    dataset_path: str | None = None,
    targets_path: str | None = None,
    compounds_path: str | None = None,
) -> dict[str, str | None]:
    """Collect artifact file paths from a run directory.

    Args:
        run_dir: Path to the run directory
        dataset_path: Optional explicit path to dataset.csv
        targets_path: Optional explicit path to targets.csv
        compounds_path: Optional explicit path to compounds.csv

    Returns:
        Dictionary mapping artifact keys to file paths (or None if not found)
    """
    run_path = Path(run_dir).resolve() if run_dir else None

    def _maybe(p: Path) -> str | None:
        try:
            return str(p) if p.exists() else None
        except Exception:
            return None

    # If caller didn't provide explicit paths, fall back to conventional names in run_dir.
    if run_path is not None:
        if dataset_path is None:
            dataset_path = _maybe(run_path / "dataset.csv")
        if targets_path is None:
            targets_path = _maybe(run_path / "targets.csv")
        if compounds_path is None:
            compounds_path = _maybe(run_path / "compounds.csv")

        metadata_path = _maybe(run_path / "metadata.json")

        # Model artifacts (RandomForest baseline)
        model_path = _maybe(run_path / "model_rf.pkl")
        model_metrics_path = _maybe(run_path / "model_metrics.json")
        model_predictions_path = _maybe(run_path / "model_predictions.csv")

        # Model artifacts (GNN)
        gnn_model_path = _maybe(run_path / "model_gnn.pt")
        gnn_metrics_path = _maybe(run_path / "model_metrics_gnn.json")
        gnn_predictions_path = _maybe(run_path / "model_predictions_gnn.csv")

        encoder_path = _maybe(run_path / "encoder_gnn.pt")
        encoder_config_path = _maybe(run_path / "encoder_gnn_config.json")
        molecule_embeddings_path = _maybe(run_path / "molecule_embeddings.npz")

        molecule_features_path = _maybe(run_path / "molecule_features.csv")
        protein_features_path = _maybe(run_path / "protein_features.csv")

        zip_path = str(run_path / "artifacts.zip")
    else:
        metadata_path = None
        model_path = None
        model_metrics_path = None
        model_predictions_path = None
        gnn_model_path = None
        gnn_metrics_path = None
        gnn_predictions_path = None
        encoder_path = None
        encoder_config_path = None
        molecule_embeddings_path = None
        molecule_features_path = None
        protein_features_path = None

        zip_path = None

    return {
        "dataset": dataset_path,
        "targets": targets_path,
        "compounds": compounds_path,
        "metadata": metadata_path,
        "model": model_path,
        "model_metrics": model_metrics_path,
        "model_predictions": model_predictions_path,
        "model_gnn": gnn_model_path,
        "model_metrics_gnn": gnn_metrics_path,
        "model_predictions_gnn": gnn_predictions_path,
        "encoder_gnn": encoder_path,
        "encoder_gnn_config": encoder_config_path,
        "molecule_embeddings": molecule_embeddings_path,
        "molecule_features": molecule_features_path,
        "protein_features": protein_features_path,
        "zip": zip_path,
    }


def write_artifacts_zip(
    *, zip_path: str | None, paths: list[str | None]
) -> str | None:
    """Create a zip file from a list of artifact paths.

    Args:
        zip_path: Path where the zip file should be created
        paths: List of file paths to include in the zip

    Returns:
        Path to the created zip file, or None if creation failed
    """
    if not zip_path:
        return None
    try:
        zpath = Path(zip_path)
        zpath.parent.mkdir(parents=True, exist_ok=True)
        with zipfile.ZipFile(zpath, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            for p in paths:
                if p and os.path.exists(p):
                    zf.write(p, arcname=os.path.basename(p))
        return str(zpath)
    except Exception:
        return None


def write_artifacts_zip_from_manifest(
    *, artifacts: dict[str, str | None]
) -> str | None:
    """Create a zip file from an artifacts manifest dictionary.

    Args:
        artifacts: Dictionary mapping artifact keys to file paths

    Returns:
        Path to the created zip file, or None if creation failed
    """
    zip_path = artifacts.get("zip")
    keys = artifact_keys_in_zip()
    paths = [artifacts.get(k) for k in keys]
    return write_artifacts_zip(zip_path=zip_path, paths=paths)


def artifacts_table(artifacts: dict[str, str | None]) -> pd.DataFrame:
    """Create a DataFrame table of artifacts for UI display.

    Args:
        artifacts: Dictionary mapping artifact keys to file paths

    Returns:
        DataFrame with columns 'artifact' and 'path'
    """
    return pd.DataFrame(
        [
            {"artifact": "dataset.csv", "path": artifacts.get("dataset") or ""},
            {"artifact": "targets.csv", "path": artifacts.get("targets") or ""},
            {"artifact": "compounds.csv", "path": artifacts.get("compounds") or ""},
            {"artifact": "metadata.json", "path": artifacts.get("metadata") or ""},
            {
                "artifact": "molecule_features.csv",
                "path": artifacts.get("molecule_features") or "",
            },
            {"artifact": "model_rf.pkl", "path": artifacts.get("model") or ""},
            {
                "artifact": "model_metrics.json",
                "path": artifacts.get("model_metrics") or "",
            },
            {
                "artifact": "model_predictions.csv",
                "path": artifacts.get("model_predictions") or "",
            },
            {"artifact": "model_gnn.pt", "path": artifacts.get("model_gnn") or ""},
            {
                "artifact": "model_metrics_gnn.json",
                "path": artifacts.get("model_metrics_gnn") or "",
            },
            {
                "artifact": "model_predictions_gnn.csv",
                "path": artifacts.get("model_predictions_gnn") or "",
            },
            {"artifact": "encoder_gnn.pt", "path": artifacts.get("encoder_gnn") or ""},
            {
                "artifact": "encoder_gnn_config.json",
                "path": artifacts.get("encoder_gnn_config") or "",
            },
            {
                "artifact": "molecule_embeddings.npz",
                "path": artifacts.get("molecule_embeddings") or "",
            },
            {"artifact": "artifacts.zip", "path": artifacts.get("zip") or ""},
        ]
    )




import pytest
from unittest.mock import MagicMock, patch
import pandas as pd

# Mocking gradio before importing ui avoids launching any web server stuff
with patch.dict("sys.modules", {"gradio": MagicMock()}):
    from dta_gnn.app.ui import (
        _update_model_choices,
        _sync_ho_gnn_arch,
        _sync_ho_wandb_section,
        build_ui,
    )

def test_sync_ho_gnn_arch():
    """Test logic for showing/hiding GNN arch params."""
    # Should hide if model != GNN
    update = _sync_ho_gnn_arch("gin", "RandomForest")
    # Since we mocked gradio, gr.update() returns a Mock. 
    # We can check if it was called (or just that the function runs without error).
    # Realistically, we'd check the properties of the returned update object if it wasn't a Mock.
    assert update is not None

def test_sync_ho_wandb_section():
    """Test logic for showing W&B section."""
    update_visible = _sync_ho_wandb_section("W&B Sweeps (Bayes)", "GNN")
    # We assume the mock returns something truthy or we can inspect calls if we mocked gr.update explicitly
    assert update_visible is not None

@patch("dta_gnn.app.ui.gr")
@patch("dta_gnn.app.ui.find_chembl_sqlite_dbs")
def test_build_ui_smoke(mock_find_dbs, mock_gr):
    """Smoke test for build_ui. Verifies it constructs without error."""
    mock_find_dbs.return_value = ["db1.sqlite"]
    
    # Mock all the component constructors to avoid side effects
    mock_gr.Blocks.return_value.__enter__.return_value = MagicMock()
    
    # Call build_ui
    try:
        demo, components = build_ui()
    except Exception as e:
        pytest.fail(f"build_ui() raised exception: {e}")
    
    # Check that it returned a tuple
    assert demo is not None
    assert components is not None



import pandas as pd
import time
import random
from typing import Optional, List
from chembl_webresource_client.new_client import new_client
from loguru import logger
from dta_gnn.io.chembl_source import ChemblSource


# Retry configuration
MAX_RETRIES = 5
BASE_DELAY = 2  # seconds
MAX_DELAY = 60  # seconds


def retry_with_backoff(func, max_retries=MAX_RETRIES, base_delay=BASE_DELAY, max_delay=MAX_DELAY):
    """Execute a function with exponential backoff retry logic."""
    last_exception = None
    for attempt in range(max_retries):
        try:
            return func()
        except Exception as e:
            last_exception = e
            error_str = str(e).lower()
            # Check if it's a server error (500, 502, 503, 504) or connection error
            is_retryable = any(
                x in error_str for x in 
                ["500", "502", "503", "504", "server", "connection", "timeout", "temporary"]
            )
            
            if not is_retryable or attempt == max_retries - 1:
                raise
            
            # Exponential backoff with jitter
            delay = min(base_delay * (2 ** attempt) + random.uniform(0, 1), max_delay)
            logger.warning(
                f"ChEMBL API error (attempt {attempt + 1}/{max_retries}): {str(e)[:100]}. "
                f"Retrying in {delay:.1f}s..."
            )
            time.sleep(delay)
    
    raise last_exception


class ChemblWebSource(ChemblSource):
    """ChEMBL data source using the official Web Resource Client."""

    def fetch_activities(
        self,
        target_ids: Optional[List[str]] = None,
        molecule_ids: Optional[List[str]] = None,
        standard_types: Optional[List[str]] = None,
        progress_callback: Optional[callable] = None,
    ) -> pd.DataFrame:
        activity = new_client.activity
        query = activity

        if target_ids:
            query = query.filter(target_chembl_id__in=target_ids)
        if molecule_ids:
            query = query.filter(molecule_chembl_id__in=molecule_ids)
        if standard_types:
            query = query.filter(standard_type__in=standard_types)

        # Only human? Or configurable? For now, we fetch, user filters later or we add arg.
        # Let's add simple filtering to valid values
        query = query.filter(standard_value__isnull=False)

        # Iterate and convert to list (client returns lazy query set)
        # Warning: This can be slow for large datasets
        
        # Get total count if possible for better logging (with retry)
        try:
            total_count = retry_with_backoff(lambda: query.count())
            logger.info(f"Fetching {total_count} activities from Web...")
            if progress_callback:
                progress_callback(0, total_count, "Starting fetch...")
        except Exception as e:
            logger.warning(f"Could not get activity count: {e}")
            total_count = None
            logger.info("Fetching activities from Web...")
            if progress_callback:
                progress_callback(0, None, "Starting fetch...")

        records = []
        count = 0
        last_log_time = time.time()
        consecutive_errors = 0
        max_consecutive_errors = 3

        # Use iterator with retry logic
        query_iter = iter(query)
        while True:
            try:
                act = retry_with_backoff(lambda: next(query_iter))
                consecutive_errors = 0  # Reset on success
            except StopIteration:
                break
            except Exception as e:
                consecutive_errors += 1
                error_msg = str(e)
                if "500" in error_msg or "server" in error_msg.lower():
                    raise RuntimeError(
                        f"ChEMBL API is experiencing server issues (HTTP 500). "
                        f"This is a temporary problem on their end. "
                        f"Please try again in a few minutes. "
                        f"Fetched {count} records before failure."
                    ) from e
                if consecutive_errors >= max_consecutive_errors:
                    raise RuntimeError(
                        f"Failed to fetch activities after {consecutive_errors} consecutive errors. "
                        f"Last error: {error_msg}. Fetched {count} records before failure."
                    ) from e
                logger.warning(f"Error fetching activity, retrying: {error_msg[:100]}")
                continue

            records.append(
                {
                    "molecule_chembl_id": act["molecule_chembl_id"],
                    "target_chembl_id": act["target_chembl_id"],
                    "standard_type": act["standard_type"],
                    "standard_value": act["standard_value"],
                    "standard_units": act["standard_units"],
                    "standard_relation": act.get("standard_relation", "="),
                    "pchembl_value": act.get("pchembl_value"),
                    "year": act.get("document_year"),
                }
            )
            count += 1
            
            # Log progress every 2 seconds or every 500 records
            # We want frequent enough updates for UI but not spam
            if count % 100 == 0 or (time.time() - last_log_time) > 1.0:
                last_log_time = time.time()
                if total_count:
                    percent = (count / total_count) * 100
                    logger.info(f"Fetched {count}/{total_count} activities ({percent:.1f}%)")
                    if progress_callback:
                        progress_callback(count, total_count, f"Fetched {count}/{total_count}")
                else:
                    logger.info(f"Fetched {count} activities")
                    if progress_callback:
                        progress_callback(count, None, f"Fetched {count}")

        logger.info(f"Completed fetching {len(records)} activities from Web.")
        if progress_callback:
            progress_callback(len(records), total_count, "Done fetching.")

        return pd.DataFrame(records)

    def fetch_molecules(self, molecule_ids: List[str]) -> pd.DataFrame:
        molecule = new_client.molecule
        # Chunking might be needed if list is too long
        query = molecule.filter(molecule_chembl_id__in=molecule_ids).only(
            ["molecule_chembl_id", "molecule_structures"]
        )

        data = []
        query_iter = iter(query)
        consecutive_errors = 0
        
        while True:
            try:
                m = retry_with_backoff(lambda: next(query_iter))
                consecutive_errors = 0
            except StopIteration:
                break
            except Exception as e:
                consecutive_errors += 1
                if consecutive_errors >= 3:
                    raise RuntimeError(
                        f"Failed to fetch molecules after multiple errors. "
                        f"ChEMBL API may be experiencing issues. Error: {str(e)[:100]}"
                    ) from e
                continue
                
            smiles = (
                m["molecule_structures"]["canonical_smiles"]
                if m["molecule_structures"]
                else None
            )
            data.append(
                {"molecule_chembl_id": m["molecule_chembl_id"], "smiles": smiles}
            )
        return pd.DataFrame(data)

    def fetch_targets(self, target_ids: List[str]) -> pd.DataFrame:
        target = new_client.target
        query = target.filter(target_chembl_id__in=target_ids)

        data = []
        query_iter = iter(query)
        consecutive_errors = 0
        
        while True:
            try:
                t = retry_with_backoff(lambda: next(query_iter))
                consecutive_errors = 0
            except StopIteration:
                break
            except Exception as e:
                consecutive_errors += 1
                if consecutive_errors >= 3:
                    raise RuntimeError(
                        f"Failed to fetch targets after multiple errors. "
                        f"ChEMBL API may be experiencing issues. Error: {str(e)[:100]}"
                    ) from e
                continue

            # We specifically want protein sequences, which are in components
            # This is a simplification; ChEMBL targets can be complex.
            # We take the first component's sequence if available.
            if "target_components" in t and t["target_components"]:
                # Accessing components details might require separate fetch in some client versions,
                # but let's assume standard object structure.
                # Actually, usually target details have component links.
                pass

            # Simple fallback: use target_chembl_id to fetch component sequences
            # This part is tricky with web client.
            # For simplicity in this implementation step, we stick to basic target info
            # Users often match by Uniprot ID.

            data.append(
                {
                    "target_chembl_id": t["target_chembl_id"],
                    "organism": t.get("organism"),
                    "target_type": t.get("target_type"),
                }
            )
        return pd.DataFrame(data)

    def get_targets(self, accession: Optional[str] = None) -> List[dict]:
        """Get targets by UniProt accession.

        Args:
            accession: UniProt accession (e.g., 'P00533' for EGFR)

        Returns:
            List of target dictionaries with target_chembl_id
        """
        target = new_client.target

        if accession:
            # Search for targets containing this accession in their components
            try:
                # The ChEMBL API allows filtering by component accession
                records = retry_with_backoff(
                    lambda: list(
                        target.filter(target_components__accession=accession).only(
                            ["target_chembl_id", "pref_name", "organism"]
                        )
                    )
                )
                return records
            except Exception:
                # Fallback: try searching via target_component endpoint
                try:
                    target_component = new_client.target_component
                    components = retry_with_backoff(
                        lambda: list(target_component.filter(accession=accession))
                    )

                    # Get unique target IDs from components
                    target_ids = set()
                    for comp in components:
                        if "targets" in comp:
                            for t in comp["targets"]:
                                if "target_chembl_id" in t:
                                    target_ids.add(t["target_chembl_id"])

                    if target_ids:
                        return [{"target_chembl_id": tid} for tid in target_ids]
                except Exception:
                    pass

                return []

        return []



"""Model utility functions for listing and managing trained models."""

from __future__ import annotations

import random
from pathlib import Path

import numpy as np

from dta_gnn.io.runs import resolve_current_run_dir


def set_seed(seed: int = 42) -> None:
    """Set random seeds for reproducibility.

    Seeds:
    - random
    - numpy
    - torch (if available)
    - torch.cuda (if available)
    - torch.backends.cudnn (deterministic)
    """
    random.seed(seed)
    np.random.seed(seed)

    try:
        import torch

        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False
    except ImportError:
        pass


def list_available_models(run_dir: Path | None = None) -> dict[str, list[str]]:
    """List all available trained models in the run directory.

    Args:
        run_dir: Path to run directory. If None, attempts to resolve current run directory.

    Returns:
        Dictionary with keys: 'rf', 'svr', 'gnn'
        Each value is a list of model identifiers (for GNN: architecture names)
    """
    if run_dir is None:
        try:
            run_dir = resolve_current_run_dir()
        except FileNotFoundError:
            return {"rf": [], "svr": [], "gnn": []}

    if not run_dir or not run_dir.exists():
        return {"rf": [], "svr": [], "gnn": []}

    models = {"rf": [], "svr": [], "gnn": []}

    # Check for RandomForest model
    rf_model = run_dir / "model_rf.pkl"
    if rf_model.exists():
        models["rf"].append("RandomForest")

    # Check for SVR model
    svr_model = run_dir / "model_svr.pkl"
    if svr_model.exists():
        models["svr"].append("SVR")

    # Check for GNN models (model_gnn_<architecture>.pt)
    gnn_architectures = [
        "gin",
        "gcn",
        "gat",
        "sage",
        "pna",
        "transformer",
        "tag",
        "arma",
        "cheb",
        "supergat",
    ]
    for arch in gnn_architectures:
        gnn_model = run_dir / f"model_gnn_{arch}.pt"
        config_file = run_dir / f"encoder_{arch}_config.json"
        
        # Correction in previous write attempt: variable name typo fix
        if gnn_model.exists() and config_file.exists():
            # Format: "GNN (GIN)", "GNN (Transformer)", etc.
            arch_display = (
                arch.upper() if arch in ["gin", "gat", "pna"] else arch.capitalize()
            )
            models["gnn"].append(f"GNN ({arch_display})")

    return models



import typer
from pathlib import Path

app = typer.Typer(help="DTA-GNN CLI")


@app.command()
def audit(
    file: Path = typer.Argument(..., help="Path to dataset CSV"),
):
    """
    Run audit on an existing dataset file.
    """
    # todo: load and audit
    typer.echo("Audit completed (Mock).")


@app.command()
def setup(
    version: str = typer.Option("36", help="ChEMBL version to download"),
    dir: Path = typer.Option(".", help="Directory to download to"),
):
    """
    Download and set up the ChEMBL SQLite database.
    """
    from dta_gnn.io.downloader import download_chembl_db

    typer.echo(f"Downloading ChEMBL {version} to {dir}...")
    try:
        db_path = download_chembl_db(version, str(dir))
        typer.echo(f"Successfully set up database at: {db_path}")
        typer.echo("You can now use this path with --db-path or in the UI.")
    except Exception as e:
        typer.echo(f"Setup failed: {e}", err=True)


@app.command()
def ui(
    host: str = typer.Option(
        "127.0.0.1", "--host", "-h", help="Host to bind to. Use 0.0.0.0 for Docker."
    ),
    port: int = typer.Option(7860, "--port", "-p", help="Port to run the server on."),
    share: bool = typer.Option(False, "--share", help="Create a public Gradio link."),
):
    """
    Launch the Gradio UI.
    """
    from dta_gnn.app.ui import launch

    launch(host=host, port=port, share=share)


if __name__ == "__main__":
    app()



"""Tests for CLI commands."""

from typer.testing import CliRunner

from dta_gnn.cli import app


runner = CliRunner()


class TestCli:
    """Tests for the CLI application."""

    def test_help(self):
        """Test that help command works."""
        result = runner.invoke(app, ["--help"])

        assert result.exit_code == 0
        assert "DTA-GNN" in result.stdout

    def test_audit_help(self):
        """Test audit subcommand help."""
        result = runner.invoke(app, ["audit", "--help"])

        assert result.exit_code == 0

    def test_ui_help(self):
        """Test ui subcommand help."""
        result = runner.invoke(app, ["ui", "--help"])

        assert result.exit_code == 0



"""Model training, evaluation, and inference utilities."""

from .gnn import (
    GnnEmbeddingExtractResult,
    GnnTrainConfig,
    GnnTrainResult,
    extract_gnn_embeddings_on_run,
    train_gnn_on_run,
)
from .hyperopt import (
    HyperoptConfig,
    HyperoptResult,
    optimize_gnn,
    optimize_gnn_wandb,
    optimize_random_forest,
    optimize_random_forest_wandb,
    optimize_svr_wandb,
)
from .predict import (
    PredictionResult,
    predict_with_gnn,
    predict_with_random_forest,
    predict_with_svr,
)
from .random_forest import train_random_forest_on_run
from .svr import train_svr_on_run
from .utils import list_available_models  # noqa: F401

__all__ = [
    "train_random_forest_on_run",
    "train_svr_on_run",
    "GnnTrainConfig",
    "GnnTrainResult",
    "GnnEmbeddingExtractResult",
    "train_gnn_on_run",
    "extract_gnn_embeddings_on_run",
    "HyperoptConfig",
    "HyperoptResult",
    "optimize_random_forest",
    "optimize_random_forest_wandb",
    "optimize_svr_wandb",
    "optimize_gnn",
    "optimize_gnn_wandb",
    "PredictionResult",
    "predict_with_random_forest",
    "predict_with_svr",
    "predict_with_gnn",
    "list_available_models",
]




import pytest
import sys
from unittest.mock import MagicMock, patch
import pandas as pd
import numpy as np
from pathlib import Path
from dta_gnn.models.predict import predict_with_random_forest, predict_with_svr, predict_with_gnn

@pytest.fixture
def mock_joblib():
    return MagicMock()

@pytest.fixture
def mock_rdkit():
    mock = MagicMock()
    mock.Chem.MolFromSmiles.return_value = MagicMock()
    mock.AllChem.GetMorganFingerprintAsBitVect.return_value = [0]*2048
    return mock

@pytest.fixture
def mock_torch_geom():
    mock = MagicMock()
    return mock

def test_predict_with_random_forest(mock_joblib, mock_rdkit):
    """Test predict_with_random_forest logic."""
    
    # We patch sys.modules via patch.dict wrapper
    with patch.dict("sys.modules", {
        "joblib": mock_joblib,
        "rdkit": mock_rdkit,
        "rdkit.Chem": mock_rdkit.Chem,
        "rdkit.Chem.AllChem": mock_rdkit.AllChem,
    }):
        # Mock Path to return true for model existence
        with patch("dta_gnn.models.predict.Path") as mock_path:
            mock_run_dir = MagicMock()
            mock_path.return_value.resolve.return_value = mock_run_dir
            mock_model_path = mock_run_dir / "model_rf.pkl"
            mock_model_path.exists.return_value = True
            
            # Mock Model
            mock_model = MagicMock()
            mock_joblib.load.return_value = mock_model
            # Set up prediction returns
            mock_model.predict.return_value = np.array([0.5, 0.8])
            mock_model.predict_proba.return_value = np.array([[0.5, 0.5], [0.2, 0.8]])
            
            smiles_list = ["C", "CC"]
            
            result = predict_with_random_forest(
                run_dir="output/run_1",
                smiles_list=smiles_list
            )
            
            assert result.model_type == "RandomForest"
            assert len(result.predictions) == 2
            assert "prediction" in result.predictions.columns
            assert "molecule_id" in result.predictions.columns
            assert result.predictions.iloc[0]["molecule_id"] == "mol_0"

def test_predict_with_random_forest_file_not_found():
    """Test error when model file is missing."""
    with patch("dta_gnn.models.predict.Path") as mock_path:
        mock_run_dir = MagicMock()
        mock_path.return_value.resolve.return_value = mock_run_dir
        (mock_run_dir / "model_rf.pkl").exists.return_value = False
        
        # Don't strictly need to patch libraries if it fails earlier
        with pytest.raises(FileNotFoundError):
            predict_with_random_forest("output/run_1", ["C"])

def test_predict_with_svr(mock_joblib, mock_rdkit):
    """Test predict_with_svr logic."""
    
    with patch.dict("sys.modules", {
        "joblib": mock_joblib,
        "rdkit": mock_rdkit,
        "rdkit.Chem": mock_rdkit.Chem,
        "rdkit.Chem.AllChem": mock_rdkit.AllChem,
        "rdkit.DataStructs": mock_rdkit.DataStructs,
    }):
        with patch("dta_gnn.models.predict.Path") as mock_path:
            mock_run_dir = MagicMock()
            mock_path.return_value.resolve.return_value = mock_run_dir
            mock_model_path = mock_run_dir / "model_svr.pkl"
            mock_model_path.exists.return_value = True
            
            mock_model = MagicMock()
            mock_joblib.load.return_value = mock_model
            # SVR predict returns array directly
            mock_model.predict.return_value = np.array([5.5, 6.0])
            
            smiles_list = ["C", "CC"]
            result = predict_with_svr(
                run_dir="output/run_1",
                smiles_list=smiles_list
            )
            
            assert result.model_type == "SVR"
            assert len(result.predictions) == 2
            assert result.predictions.iloc[0]["prediction"] == 5.5

def test_predict_with_gnn(mock_joblib, mock_rdkit, mock_torch_geom):
    """Test predict_with_gnn logic, including GNN architecture detection."""
    
    # Mock torch
    mock_torch = MagicMock()
    mock_torch.load.return_value = {} # Empty state dict
    mock_torch.no_grad.return_value.__enter__.return_value = None
    
    # Mock PyG
    mock_data = MagicMock()
    mock_loader = MagicMock()
    
    # Setup mock loader iteration
    # Iterate once, returning a batch
    mock_batch = MagicMock()
    mock_batch.to.return_value = mock_batch # .to(device)
    mock_batch.molecule_chembl_id = ["mol1", "mol2"]
    mock_loader.return_value = [mock_batch]
    
    with patch.dict("sys.modules", {
        "torch": mock_torch,
        "torch_geometric": mock_torch_geom,
        "torch_geometric.data": MagicMock(Data=mock_data),
        "torch_geometric.loader": MagicMock(DataLoader=mock_loader),
    }):
        # Mock internal dependencies explicitly
        # We patch where they are IMPORTED in predict.py scope? 
        # predict.py: from dta_gnn.features.molecule_graphs import smiles_to_graph_2d
        # If we didn't patch dta_gnn.features..., it imports the REAL one.
        # The real one imports torch. Torch is in sys.modules as MOCK.
        # So real module loads with Mock torch. That should be OK.
        
        # We patch the function on the real module (or whatever is loaded)
        # But we need to make sure the import in predict.py sees our mock, or we patch the name in place.
        # Since imports are inside the function, patching the source module works.
        
        with patch("dta_gnn.features.molecule_graphs.smiles_to_graph_2d") as mock_s2g, \
             patch("dta_gnn.models.gnn._make_encoder_and_model") as mock_make_model, \
             patch("dta_gnn.models.gnn._get_device") as mock_get_device, \
             patch("dta_gnn.models.gnn._load_json") as mock_load_json, \
             patch("dta_gnn.models.predict.Path") as mock_path:
                
                # Setup Paths
                mock_run_dir = MagicMock()
                mock_path.return_value.resolve.return_value = mock_run_dir
                
                # Case: Explicit architecture provided
                mock_model_path = mock_run_dir / "model_gnn_gin.pt"
                mock_config_path = mock_run_dir / "encoder_gin_config.json"
                mock_model_path.exists.return_value = True
                mock_config_path.exists.return_value = True
                
                # Mock config loading
                mock_json_file = MagicMock()
                mock_json_file.__enter__.return_value = mock_json_file
                # We need to mock open() as well since json.load reads from file object
                # But internal logic calls open() on Path object if possible, or python open()
                # predict.py uses `with open(config_path, "r") as f: config = json.load(f)`
                
                # Easier to mock json.load if we patch open globally
                with patch("builtins.open", MagicMock()) as mock_open, \
                     patch("json.load") as mock_json_load:
                    
                    mock_json_load.return_value = {"encoder": {"architecture": "gin"}}
                    
                    # Mock Model Construction
                    mock_predictor_cls = MagicMock()
                    mock_predictor_instance = MagicMock()
                    mock_make_model.return_value = (None, mock_predictor_cls)
                    mock_predictor_cls.return_value = mock_predictor_instance
                    
                    # Mock Model Prediction output: logits, embeddings
                    # Forward returns (logits, embeddings)
                    mock_logits = MagicMock()
                    # trace: logits.view(-1).float().cpu().numpy().tolist()
                    mock_logits.view.return_value.float.return_value.cpu.return_value.numpy.return_value.tolist.return_value = [0.5, 0.6]
                    
                    mock_predictor_instance.return_value = (
                        mock_logits, 
                        None
                    )
                    
                    # Mock Graph Conversion
                    mock_graph = MagicMock()
                    mock_graph.atom_type = np.array([0])
                    mock_graph.atom_feat = np.array([[0.1]])
                    mock_graph.edge_index = np.array([[0], [0]])
                    mock_graph.edge_attr = np.array([[0.1]])
                    mock_s2g.return_value = mock_graph
                    
                    # Run Test
                    result = predict_with_gnn(
                        run_dir="output/run_1",
                        smiles_list=["C", "CC"], 
                        architecture="gin"
                    )
                    
                    assert result.model_type == "GNN"
                    assert len(result.predictions) == 2





"""Tests for app feature helpers (compound and protein processing)."""

import pandas as pd
import pytest

from dta_gnn.app_features.compound import (
    build_smiles_frame,
    featurize_smiles_morgan,
    _parse_smiles_text,
)
from dta_gnn.app_features.proteins import (
    build_sequence_frame,
    _parse_sequences_text,
)


class TestParseSmilesText:
    """Tests for SMILES text parsing."""

    def test_comma_separated(self):
        """Test parsing comma-separated SMILES."""
        result = _parse_smiles_text("CCO, CC(=O)O, c1ccccc1")
        assert result == ["CCO", "CC(=O)O", "c1ccccc1"]

    def test_newline_separated(self):
        """Test parsing newline-separated SMILES."""
        result = _parse_smiles_text("CCO\nCC(=O)O\nc1ccccc1")
        assert result == ["CCO", "CC(=O)O", "c1ccccc1"]

    def test_mixed_separators(self):
        """Test parsing with mixed separators."""
        result = _parse_smiles_text("CCO, CC(=O)O\nc1ccccc1")
        assert len(result) == 3

    def test_empty_string(self):
        """Test parsing empty string."""
        result = _parse_smiles_text("")
        assert result == []

    def test_whitespace_stripped(self):
        """Test that whitespace is stripped."""
        result = _parse_smiles_text("  CCO  ,  CC(=O)O  ")
        assert result == ["CCO", "CC(=O)O"]


class TestBuildSmilesFrame:
    """Tests for build_smiles_frame function."""

    def test_from_text_input(self):
        """Test building frame from text input."""
        result = build_smiles_frame(
            smiles_text="CCO, CC(=O)O", df_state=None, source_mode="text"
        )

        assert len(result) == 2
        assert "smiles" in result.columns
        assert "molecule_chembl_id" in result.columns

    def test_from_dataset_state(self):
        """Test building frame from dataset state."""
        df_state = pd.DataFrame(
            {"molecule_chembl_id": ["M1", "M2"], "smiles": ["CCO", "CC(=O)O"]}
        )

        result = build_smiles_frame(
            smiles_text="", df_state=df_state, source_mode="dataset"
        )

        assert len(result) == 2
        assert result.loc[0, "smiles"] == "CCO"

    def test_canonical_smiles_column(self):
        """Test handling of canonical_smiles column."""
        df_state = pd.DataFrame(
            {"molecule_chembl_id": ["M1"], "canonical_smiles": ["CCO"]}
        )

        result = build_smiles_frame(
            smiles_text="", df_state=df_state, source_mode="dataset"
        )

        assert "smiles" in result.columns

    def test_empty_text_and_no_state_raises(self):
        """Test that empty text with no state raises error."""
        with pytest.raises(ValueError):
            build_smiles_frame(smiles_text="", df_state=None, source_mode="text")

    def test_generates_custom_ids(self):
        """Test that custom IDs are generated for text input."""
        result = build_smiles_frame(
            smiles_text="CCO, CC(=O)O", df_state=None, source_mode="text"
        )

        assert result.loc[0, "molecule_chembl_id"] == "custom_1"
        assert result.loc[1, "molecule_chembl_id"] == "custom_2"


class TestFeaturizeSmilessMorgan:
    """Tests for Morgan fingerprint featurization."""

    def test_basic_featurization(self):
        """Test basic Morgan fingerprint calculation."""
        df = pd.DataFrame(
            {"molecule_chembl_id": ["M1", "M2"], "smiles": ["CCO", "c1ccccc1"]}
        )

        result = featurize_smiles_morgan(df)

        assert "morgan_fingerprint" in result.columns
        assert len(result.loc[0, "morgan_fingerprint"]) == 2048

    def test_custom_parameters(self):
        """Test with custom radius and n_bits."""
        df = pd.DataFrame({"smiles": ["CCO"]})

        result = featurize_smiles_morgan(df, radius=3, n_bits=1024)

        assert len(result.loc[0, "morgan_fingerprint"]) == 1024

    def test_missing_smiles_column_raises(self):
        """Test that missing smiles column raises error."""
        df = pd.DataFrame({"other": ["value"]})

        with pytest.raises(ValueError, match="smiles"):
            featurize_smiles_morgan(df)


class TestParseSequencesText:
    """Tests for sequence text parsing."""

    def test_comma_separated(self):
        """Test parsing comma-separated sequences."""
        result = _parse_sequences_text("ACDE, FGHI, KLMN")
        assert result == ["ACDE", "FGHI", "KLMN"]

    def test_converts_to_uppercase(self):
        """Test that sequences are converted to uppercase."""
        result = _parse_sequences_text("acde, FgHi")
        assert result == ["ACDE", "FGHI"]


class TestBuildSequenceFrame:
    """Tests for build_sequence_frame function."""

    def test_from_text_input(self):
        """Test building frame from text input."""
        result = build_sequence_frame(
            seqs_text="ACDE, FGHI", df_state=None, source_mode="text"
        )

        assert len(result) == 2
        assert "sequence" in result.columns
        assert "target_chembl_id" in result.columns

    def test_from_dataset_state(self):
        """Test building frame from dataset state."""
        df_state = pd.DataFrame(
            {"target_chembl_id": ["T1", "T2"], "sequence": ["ACDE", "FGHI"]}
        )

        result = build_sequence_frame(
            seqs_text="", df_state=df_state, source_mode="dataset"
        )

        assert len(result) == 2

    def test_empty_raises(self):
        """Test that empty input raises error."""
        with pytest.raises(ValueError):
            build_sequence_frame(seqs_text="", df_state=None, source_mode="text")



from __future__ import annotations

import re

import pandas as pd


def _parse_sequences_text(seqs_text: str) -> list[str]:
    # Allow newline/comma/whitespace separated sequences.
    tokens = re.split(r"[,\n\s]+", seqs_text or "")
    seqs = [t.strip().upper() for t in tokens if t and t.strip()]
    return seqs


def build_sequence_frame(
    *,
    seqs_text: str,
    df_state: pd.DataFrame | None,
    source_mode: str,
) -> pd.DataFrame:
    """Build a minimal protein-sequence DataFrame for UI featurization."""

    mode = (source_mode or "").strip().lower()
    want_dataset = ("dataset" in mode) or ("state" in mode)

    if (not want_dataset) and seqs_text:
        seqs = _parse_sequences_text(seqs_text)
        if not seqs:
            raise ValueError("No sequences found in the provided text.")
        return pd.DataFrame(
            {
                "target_chembl_id": [f"custom_{i+1}" for i in range(len(seqs))],
                "sequence": seqs,
            }
        )

    if df_state is None or df_state.empty:
        if seqs_text:
            seqs = _parse_sequences_text(seqs_text)
            if not seqs:
                raise ValueError("No sequences found in the provided text.")
            return pd.DataFrame(
                {
                    "target_chembl_id": [f"custom_{i+1}" for i in range(len(seqs))],
                    "sequence": seqs,
                }
            )
        raise ValueError("No dataset state available and no sequences text provided.")

    df = df_state.copy()
    if "sequence" not in df.columns:
        raise ValueError("Dataset state is missing a 'sequence' column.")

    id_col = "target_chembl_id" if "target_chembl_id" in df.columns else None

    out = pd.DataFrame({"sequence": df["sequence"].astype(str).str.upper()})
    if id_col:
        out["target_chembl_id"] = df[id_col].astype(str)
    else:
        out["target_chembl_id"] = [f"row_{i+1}" for i in range(len(out))]

    out = out.replace({"sequence": {"nan": None, "None": None}})
    out = out.dropna(subset=["sequence"])
    out = out.drop_duplicates(subset=["target_chembl_id", "sequence"])
    return out



import pandas as pd
import numpy as np
from rdkit import Chem
from typing import Optional, Literal


def standardize_activities(
    df: pd.DataFrame, convert_to_pchembl: bool = True, drop_censored: bool = False
) -> pd.DataFrame:
    """
    Standardize activity values.
    - Filters rows with missing standard_value.
    - Converts nanomolar units to molar if needed (though ChEMBL standard_value is usually nM).
    - Calculates pChEMBL if missing and requested.
    """
    df = df.copy()

    # Ensure numeric columns are actually numeric
    df["standard_value"] = pd.to_numeric(df["standard_value"], errors="coerce")
    if "pchembl_value" in df.columns:
        df["pchembl_value"] = pd.to_numeric(df["pchembl_value"], errors="coerce")

    # Drop missing values
    df = df.dropna(subset=["standard_value", "standard_units"])

    # Handle censored values (>, <)
    # If standard_relation exists, we can use it.
    if "standard_relation" in df.columns:
        if drop_censored:
            # Keep only exact measurements
            df = df[
                df["standard_relation"].isin(["=", ""])
            ]  # Some might be empty string?
        else:
            pass  # Keep them for now, might be used for binary labeling depending on direction

    # ChEMBL standard_value is typically in nM.
    # pChEMBL = -log10(value_molar)
    # value_molar = value_nM * 1e-9

    if convert_to_pchembl and "pchembl_value" in df.columns:
        # If pchembl_value column exists but has NaNs, try to fill them
        mask = df["pchembl_value"].isna()
        # Only fill if units are nM (most common)
        nm_mask = df["standard_units"] == "nM"

        # Calculate for those missing pchembl but having nM values
        to_calc = mask & nm_mask
        if to_calc.any():
            # Avoid log(0)
            valid_val = df.loc[to_calc, "standard_value"] > 0
            df.loc[to_calc & valid_val, "pchembl_value"] = -np.log10(
                df.loc[to_calc & valid_val, "standard_value"] * 1e-9
            )

    return df


def aggregate_duplicates(
    df: pd.DataFrame,
    group_cols: list = ["molecule_chembl_id", "target_chembl_id"],
    agg_method: Literal["median", "mean", "max", "min"] = "median",
) -> pd.DataFrame:
    """
    Deduplicate measurements for the same drug-target pair.
    """
    if df.empty:
        return df

    # We aggregate pchembl_value if available, else standard_value
    target_col = "pchembl_value" if "pchembl_value" in df.columns else "standard_value"

    # Drop rows where the target column is NaN before aggregating
    df_clean = df.dropna(subset=[target_col])

    # Groupby and aggregate
    grouped = df_clean.groupby(group_cols)[target_col].agg(agg_method).reset_index()

    # If 'year' is present, we need to preserve it.
    # Since year is per-activity, and we are aggregating activities for the same Mol-Target pair,
    # we should take the MIN key (earliest year)? Or MAX?
    # In 'cold drug', we just need 'a' year. For temporal split, using the earliest year
    # ensures we don't leak future info (if a drug was published in 2018 and 2022, treating it as 2018 is safe for train).
    if "year" in df.columns:
        # We need to aggregate year as well.
        # Let's perform aggregation on multiple columns.
        agg_dict = {target_col: agg_method, "year": "min"}
        grouped = df_clean.groupby(group_cols).agg(agg_dict).reset_index()

    return grouped


def canonicalize_smiles(smiles: str) -> Optional[str]:
    """Canonicalize a single SMILES string using RDKit."""
    try:
        mol = Chem.MolFromSmiles(smiles)
        if mol:
            return Chem.MolToSmiles(mol, canonical=True, isomericSmiles=False)
    except Exception:
        pass
    return None



import pandas as pd
from typing import Optional, List, Literal, TYPE_CHECKING
import tempfile
from loguru import logger
from dta_gnn.cleaning import standardize_activities, aggregate_duplicates
from dta_gnn.splits import split_random, split_cold_drug_scaffold, split_temporal
from dta_gnn.features import calculate_morgan_fingerprints

if TYPE_CHECKING:
    from dta_gnn.io.chembl_source import ChemblSource


class Pipeline:
    def __init__(
        self,
        source_type: Literal["web", "sqlite"] = "web",
        sqlite_path: Optional[str] = None,
    ):
        if source_type == "sqlite":
            if not sqlite_path:
                raise ValueError("sqlite_path required for sqlite source")
            # Lazy import so UI can start even if the ChEMBL Web API is down.
            from dta_gnn.io.sqlite_source import ChemblSQLiteSource

            self.source = ChemblSQLiteSource(sqlite_path)
        else:
            # Lazy import to avoid importing chembl_webresource_client at module import time.
            from dta_gnn.io.web_source import ChemblWebSource

            self.source = ChemblWebSource()

    def build_dta(
        self,
        *,
        target_ids: Optional[List[str]] = None,
        molecule_ids: Optional[List[str]] = None,
        standard_types: Optional[List[str]] = None,
        split_method: str = "random",
        output_path: Optional[str] = None,
        test_size: float = 0.2,
        val_size: float = 0.1,
        split_year: int = 2022,
        featurize: bool = False,
        progress_callback: Optional[callable] = None,
    ) -> pd.DataFrame:
        """Build a DTA-style regression dataset.

        The regression label is always `pchembl_value` (after optional cleaning).
        Target sequences/metadata are stored separately in `self.last_targets_csv`.
        """

        logger.info(f"Starting DTA build for targets: {target_ids}")

        df_activities = self.source.fetch_activities(
            target_ids=target_ids,
            molecule_ids=molecule_ids,
            standard_types=standard_types,
            progress_callback=progress_callback,
        )

        if df_activities is None or len(df_activities) == 0:
            return pd.DataFrame()

        # Normalize/ensure pChEMBL.
        df_clean = standardize_activities(df_activities, convert_to_pchembl=True)
        df_agg = aggregate_duplicates(df_clean)

        dataset_df = df_agg.dropna(subset=["pchembl_value"]).copy()
        if dataset_df.empty:
            return pd.DataFrame()
        dataset_df["label"] = dataset_df["pchembl_value"]

        # Molecules
        mol_ids = dataset_df["molecule_chembl_id"].unique().tolist()
        df_mols = self.source.fetch_molecules(mol_ids)
        dataset_df = dataset_df.merge(df_mols, on="molecule_chembl_id", how="left")

        if featurize:
            dataset_df = calculate_morgan_fingerprints(
                dataset_df, radius=2, n_bits=2048
            )

        # Split
        if split_method == "random":
            dataset_df, _, _, _ = split_random(
                dataset_df, test_size=test_size, val_size=val_size
            )
        elif split_method == "scaffold":
            dataset_df = split_cold_drug_scaffold(
                dataset_df, test_size=test_size, val_size=val_size
            )
        elif split_method == "temporal":
            dataset_df = split_temporal(
                dataset_df, split_year=split_year, val_size=val_size
            )
        else:
            raise ValueError(f"Unknown split_method: {split_method}")

        # Targets metadata (saved separately).
        unique_targets = sorted(
            set(dataset_df["target_chembl_id"].dropna().astype(str).tolist())
        )
        targets_df = self.source.fetch_targets(unique_targets)
        tmp = tempfile.NamedTemporaryFile(
            prefix="targets_", suffix=".csv", delete=False
        )
        tmp.close()
        targets_df.to_csv(tmp.name, index=False)
        self.last_targets_csv = tmp.name

        if output_path:
            dataset_df.to_csv(output_path, index=False)

        return dataset_df




import pytest
from unittest.mock import MagicMock, patch
from dta_gnn.io.web_source import ChemblWebSource

@pytest.fixture
def mock_new_client():
    with patch("dta_gnn.io.web_source.new_client") as mock:
        yield mock

def test_fetch_activities(mock_new_client):
    """Test fetch_activities logic."""
    source = ChemblWebSource()
    
    # Mock activity endpoint
    mock_activity = MagicMock()
    mock_new_client.activity = mock_activity
    
    mock_record = {
        "molecule_chembl_id": "M1",
        "target_chembl_id": "T1",
        "standard_type": "IC50",
        "standard_value": "10",
        "standard_units": "nM",
        "document_year": 2020
    }
    # Mock chained filter calls
    # Make filter return the same mock object to support arbitrary chaining
    mock_activity.filter.return_value = mock_activity
    mock_activity.__iter__.return_value = [mock_record]
    
    # Run
    df = source.fetch_activities(target_ids=["T1"])
    
    assert len(df) == 1
    assert df.iloc[0]["molecule_chembl_id"] == "M1"
    assert df.iloc[0]["pchembl_value"] is None # Not provided in mock record

def test_fetch_molecules(mock_new_client):
    """Test fetch_molecules logic."""
    source = ChemblWebSource()
    
    mock_molecule = MagicMock()
    mock_new_client.molecule = mock_molecule
    
    mock_record = {
        "molecule_chembl_id": "M1",
        "molecule_structures": {"canonical_smiles": "C"}
    }
    
    mock_molecule.filter.return_value.only.return_value = [mock_record]
    
    df = source.fetch_molecules(["M1"])
    
    assert len(df) == 1
    assert df.iloc[0]["smiles"] == "C"

def test_get_targets(mock_new_client):
    """Test get_targets logic."""
    source = ChemblWebSource()
    
    mock_target = MagicMock()
    mock_new_client.target = mock_target
    
    mock_records = [{"target_chembl_id": "T1", "pref_name": "Target 1", "organism": "Human"}]
    mock_target.filter.return_value.only.return_value = mock_records
    
    result = source.get_targets(accession="P12345")
    
    assert len(result) == 1
    assert result[0]["target_chembl_id"] == "T1"



import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from rdkit.Chem.Scaffolds import MurckoScaffold
from typing import Tuple


def split_random(
    df: pd.DataFrame, test_size: float = 0.2, val_size: float = 0.1, seed: int = 42
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Random split into Train/Val/Test.
    """
    if len(df) == 0:
        # Create empty DataFrames with same dtypes
        col_list = df.columns
        dt = df.dtypes
        return (
            pd.DataFrame(columns=col_list).astype(dt),
            pd.DataFrame(columns=col_list).astype(dt),
            pd.DataFrame(columns=col_list).astype(dt),
            pd.DataFrame(columns=col_list).astype(dt),
        )
        
    train, test = train_test_split(df, test_size=test_size, random_state=seed)

    # Further split train to get val
    # val_size is relative to TOTAL, so we need to calculate fraction of REMAINING (Train)
    # remaining = 1 - test_size
    # fraction = val_size / remaining
    if val_size > 0:
        relative_val_size = val_size / (1 - test_size)
        train, val = train_test_split(
            train, test_size=relative_val_size, random_state=seed
        )
        val["split"] = "val"
    else:
        # Create empty DataFrame with same dtypes to avoid FutureWarning
        val = pd.DataFrame(columns=df.columns).astype(df.dtypes)

    # Add split column
    train["split"] = "train"
    test["split"] = "test"

    # Filter out empty DataFrames before concat to avoid FutureWarning
    dfs_to_concat = [d for d in [train, val, test] if len(d) > 0]
    if not dfs_to_concat:
        result = pd.DataFrame(columns=df.columns)
    else:
        result = pd.concat(dfs_to_concat, ignore_index=True)
    
    return result, train, val, test


def split_cold_drug_scaffold(
    df: pd.DataFrame,
    smiles_col: str = "smiles",
    test_size: float = 0.2,
    val_size: float = 0.1,
    seed: int = 42,
) -> pd.DataFrame:
    """
    Scaffold split (Cold Drug).
    """
    # 1. Generate scaffolds
    scaffolds = {}
    for idx, row in df.iterrows():
        smi = row[smiles_col]
        try:
            scaff = MurckoScaffold.MurckoScaffoldSmiles(smi)
            if scaff not in scaffolds:
                scaffolds[scaff] = []
            scaffolds[scaff].append(idx)
        except Exception:
            # Fallback for invalid/empty
            pass

    # 2. Sort scaffolds by size (to balance) or shuffle?
    # Standard practice: sort by num molecules to put rare scaffolds in test?
    # Or random shuffle scaffolds? DeepChem does random sort.

    scaffold_sets = list(scaffolds.values())

    # Deterministic shuffle of scaffold groups
    rng = np.random.default_rng(seed)
    rng.shuffle(scaffold_sets)

    train_idxs, val_idxs, test_idxs = [], [], []
    train_cutoff = len(df) * (1 - test_size - val_size)
    val_cutoff = len(df) * (1 - test_size)

    current_count = 0
    for group in scaffold_sets:
        if current_count < train_cutoff:
            train_idxs.extend(group)
        elif current_count < val_cutoff:
            val_idxs.extend(group)
        else:
            test_idxs.extend(group)
        current_count += len(group)

    df.loc[train_idxs, "split"] = "train"
    df.loc[val_idxs, "split"] = "val"
    df.loc[test_idxs, "split"] = "test"

    return df


def split_temporal(
    df: pd.DataFrame,
    year_col: str = "year",
    split_year: int = 2022,
    val_size: float = 0.1,
) -> pd.DataFrame:
    """
    Temporal split based on year.
    Train: year < split_year
    Test: year >= split_year
    Val: random subset of Train (or could be time-based if requested, but simple random of past is standard)
    """
    # Ensure year is numeric
    df = df.copy()
    df[year_col] = pd.to_numeric(df[year_col], errors="coerce")

    train_mask = df[year_col] < split_year
    test_mask = df[year_col] >= split_year

    # Split Train further into Train/Val if val_size > 0
    train_df = df[train_mask].copy()
    test_df = df[test_mask].copy()

    if val_size > 0 and len(train_df) > 0:
        try:
            tr, va = train_test_split(train_df, test_size=val_size, random_state=42)
            train_df = tr
            val_df = va
            val_df["split"] = "val"
        except ValueError:
            # Create empty DataFrame with same dtypes to avoid FutureWarning
            val_df = pd.DataFrame(columns=df.columns).astype(df.dtypes)
    else:
        # Create empty DataFrame with same dtypes to avoid FutureWarning
        val_df = pd.DataFrame(columns=df.columns).astype(df.dtypes)

    train_df["split"] = "train"
    test_df["split"] = "test"

    # Filter out empty DataFrames before concat to avoid FutureWarning
    dfs_to_concat = [d for d in [train_df, val_df, test_df] if len(d) > 0]
    if not dfs_to_concat:
        return pd.DataFrame(columns=df.columns)
    return pd.concat(dfs_to_concat, ignore_index=True)



"""File I/O utilities for CSV preview, path normalization, and database discovery."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import pandas as pd


@dataclass(frozen=True)
class CsvPreview:
    """Result of CSV preview operation."""

    df: pd.DataFrame | None
    error: str | None = None


def normalize_csv_path(path: str | None) -> str | None:
    """Normalize a CSV file path string.

    Args:
        path: Path string to normalize

    Returns:
        Normalized path string, or None if input is empty/None
    """
    if not path:
        return None
    p = str(path).strip()
    return p or None


def preview_csv_with_error(path: str | None, n: int = 50) -> CsvPreview:
    """Preview a CSV file with error handling.

    Args:
        path: Path to CSV file
        n: Number of rows to read (default: 50)

    Returns:
        CsvPreview object with DataFrame and optional error message
    """
    path = normalize_csv_path(path)
    if not path:
        return CsvPreview(df=None, error=None)
    try:
        return CsvPreview(df=pd.read_csv(path, nrows=n), error=None)
    except Exception as e:
        return CsvPreview(df=None, error=f"Could not read CSV: {e}")


def preview_csv(path: str | None, n: int = 50) -> pd.DataFrame | None:
    """Preview a CSV file (wrapper that returns only DataFrame).

    Args:
        path: Path to CSV file
        n: Number of rows to read (default: 50)

    Returns:
        DataFrame with first n rows, or None if error
    """
    return preview_csv_with_error(path, n=n).df


def iter_existing_files(paths: Iterable[str | None]) -> list[str]:
    """Filter a list of paths to only those that exist.

    Args:
        paths: Iterable of file paths (may include None)

    Returns:
        List of existing file paths
    """
    existing: list[str] = []
    for p in paths:
        if not p:
            continue
        try:
            if Path(p).exists():
                existing.append(p)
        except Exception:
            continue
    return existing


def find_chembl_sqlite_dbs() -> list[str]:
    """Find available ChEMBL SQLite DB files under a `chembl_dbs/` folder.

    Searches both the current working directory and the repo root (when running
    from source). Returns absolute paths.

    Returns:
        Sorted list of absolute paths to SQLite database files
    """
    candidates: list[Path] = []
    cwd_dir = Path.cwd() / "chembl_dbs"
    if cwd_dir.exists() and cwd_dir.is_dir():
        candidates.append(cwd_dir)

    # When running from source, try to find repo root
    # This function may be called from various locations, so we try multiple approaches
    try:
        # Try to find repo root by looking for common markers
        current_file = Path(__file__).resolve()
        # io/utils.py is at src/dta_gnn/io/utils.py, so repo root is 3 levels up
        repo_dir = current_file.parents[3]
        repo_candidate = repo_dir / "chembl_dbs"
        if repo_candidate.exists() and repo_candidate.is_dir():
            candidates.append(repo_candidate)
    except Exception:
        pass

    exts = {".db", ".sqlite", ".sqlite3"}
    found: list[Path] = []
    for base in candidates:
        for p in base.rglob("*"):
            if p.is_file() and p.suffix.lower() in exts:
                found.append(p.resolve())

    # Stable ordering for a nicer UX
    return sorted({str(p) for p in found})



import pandas as pd
from rdkit.Chem.Scaffolds import MurckoScaffold
from typing import Dict, Any


def audit_scaffold_leakage(
    train_df: pd.DataFrame, test_df: pd.DataFrame, smiles_col: str = "smiles"
) -> Dict[str, Any]:
    """
    Check if scaffolds from test set appear in train set.
    """

    def get_scaffolds(df):
        scaffs = set()
        for s in df[smiles_col].dropna():
            try:
                scaffs.add(MurckoScaffold.MurckoScaffoldSmiles(s))
            except Exception:
                pass
        return scaffs

    train_scaffolds = get_scaffolds(train_df)
    test_scaffolds = get_scaffolds(test_df)

    overlap = train_scaffolds.intersection(test_scaffolds)
    return {
        "train_scaffolds": len(train_scaffolds),
        "test_scaffolds": len(test_scaffolds),
        "overlap_count": len(overlap),
        "leakage_ratio": len(overlap) / len(test_scaffolds) if test_scaffolds else 0.0,
    }


def audit_target_leakage(
    train_df: pd.DataFrame, test_df: pd.DataFrame, target_col: str = "target_chembl_id"
) -> Dict[str, Any]:
    """
    Check exact target ID overlap.
    """
    train_targets = set(train_df[target_col].dropna())
    test_targets = set(test_df[target_col].dropna())

    overlap = train_targets.intersection(test_targets)
    return {
        "train_targets": len(train_targets),
        "test_targets": len(test_targets),
        "overlap_count": len(overlap),
        "leakage_ratio": len(overlap) / len(test_targets) if test_targets else 0.0,
    }
